"""Test suite for _labware_origin_math.py module."""

from typing import List, NamedTuple, Optional

import pytest

from opentrons_shared_data.deck import load as load_deck
from opentrons_shared_data.deck.types import DeckDefinitionV5
from opentrons_shared_data.labware.labware_definition import (
    AxisAlignedBoundingBox3D,
    Dimensions,
    Extents,
    GripperOffsets,
    LabwareDefinition,
    LabwareDefinition2,
    LabwareDefinition3,
    LabwareRole,
    Parameters2,
    Parameters3,
    Vector3D,
)
from opentrons_shared_data.labware.types import (
    FlatSupportThermalCouplingAsChildFeature,
    HeaterShakerUniversalFlatAdapterFeature,
    LocatingFeatures,
    OpentronsFlexTipRackLidAsChildFeature,
    OpentronsFlexTipRackLidAsParentFeature,
    ScrewAnchoredAsChildFeature,
    ScrewAnchoredAsParentFeature,
    SlotFootprintAsChildFeature,
    SlotFootprintAsParentFeature,
    Vector2D,
)
from opentrons_shared_data.module.types import ModuleOrientation

from opentrons.protocol_engine.state.labware_origin_math.errors import (
    InvalidLabwarePlacementError,
)
from opentrons.protocol_engine.state.labware_origin_math.stackup_origin_to_labware_origin import (
    LabwareOriginContext,
    LabwareStackupAncestorDefinition,
    get_stackup_origin_to_labware_origin,
)
from opentrons.protocol_engine.types import (
    AddressableArea,
    AddressableAreaLocation,
    AddressableOffsetVector,
    AreaType,
    DeckSlotLocation,
    LabwareLocation,
    LabwareMovementOffsetData,
    LabwareOffsetVector,
    ModuleDefinition,
    ModuleDimensions,
    ModuleLocation,
    ModuleModel,
    OnLabwareLocation,
)
from opentrons.protocol_engine.types import (
    Dimensions as AddressableAreaDimensions,
)
from opentrons.protocols.api_support.deck_type import (
    STANDARD_OT2_DECK,
    STANDARD_OT3_DECK,
)
from opentrons.types import DeckSlotName, Point

# Test fixtures for labware definitions
_LW_V2 = LabwareDefinition2.model_construct(  # type: ignore[call-arg]
    namespace="test",
    version=1,
    schemaVersion=2,
    cornerOffsetFromSlot=Vector3D(x=150, y=250, z=350),
    stackingOffsetWithModule={},
    gripperOffsets={
        "default": GripperOffsets(
            pickUpOffset=Vector3D(x=100, y=200, z=300),
            dropOffset=Vector3D(x=300, y=200, z=100),
        ),
    },
    stackingOffsetWithLabware={
        "labware-name": Vector3D(x=0, y=0, z=0),
        "default": Vector3D(x=0, y=0, z=0),
        "adapter-labware": Vector3D(x=0, y=0, z=0),
    },
    dimensions=Dimensions(xDimension=1000, yDimension=1200, zDimension=750),
    parameters=Parameters2.model_construct(loadName="labware-name"),  # type: ignore[call-arg]
)

_LW_V2_WITH_MODULE_STACKING = LabwareDefinition2.model_construct(  # type: ignore[call-arg]
    namespace="test",
    version=1,
    schemaVersion=2,
    cornerOffsetFromSlot=Vector3D(x=150, y=250, z=350),
    stackingOffsetWithModule={
        str(ModuleModel.TEMPERATURE_MODULE_V2.value): Vector3D(x=50, y=100, z=150),
        str(ModuleModel.THERMOCYCLER_MODULE_V1.value): Vector3D(x=200, y=300, z=400),
        str(ModuleModel.THERMOCYCLER_MODULE_V2.value): Vector3D(x=500, y=600, z=700),
    },
    gripperOffsets={},
)

_LW_V2_WITH_LABWARE_STACKING = LabwareDefinition2.model_construct(  # type: ignore[call-arg]
    namespace="test",
    version=1,
    schemaVersion=2,
    cornerOffsetFromSlot=Vector3D(x=200, y=300, z=400),
    stackingOffsetWithLabware={
        "labware-name": Vector3D(x=50, y=100, z=150),
        "default": Vector3D(x=250, y=350, z=450),
    },
    gripperOffsets={},
)

_LW_V2_WITH_DEFAULT_GRIPPER_OFFSETS = LabwareDefinition2.model_construct(  # type: ignore[call-arg]
    namespace="test",
    version=1,
    schemaVersion=2,
    cornerOffsetFromSlot=Vector3D(x=0, y=0, z=0),
    stackingOffsetWithModule={},
    dimensions=Dimensions(xDimension=1000, yDimension=1200, zDimension=750),
    gripperOffsets={
        "default": GripperOffsets(
            pickUpOffset=Vector3D(x=10, y=20, z=30),
            dropOffset=Vector3D(x=40, y=50, z=60),
        )
    },
    stackingOffsetWithLabware={
        "labware-name": Vector3D(x=50, y=100, z=150),
        "default": Vector3D(x=250, y=350, z=450),
    },
    parameters=Parameters2.model_construct(loadName="labware-name"),  # type: ignore[call-arg]
)

_LW_V2_WITH_SLOT_SPECIFIC_GRIPPER_OFFSETS = LabwareDefinition2.model_construct(  # type: ignore[call-arg]
    namespace="test",
    version=1,
    schemaVersion=2,
    cornerOffsetFromSlot=Vector3D(x=0, y=0, z=0),
    stackingOffsetWithLabware={"default": Vector3D(x=0, y=0, z=0)},
    stackingOffsetWithModule={},
    dimensions=Dimensions(xDimension=1000, yDimension=1200, zDimension=750),
    gripperOffsets={
        "A1": GripperOffsets(
            pickUpOffset=Vector3D(x=111, y=222, z=333),
            dropOffset=Vector3D(x=333, y=222, z=111),
        ),
    },
    parameters=Parameters2.model_construct(loadName="adapter-labware"),  # type: ignore[call-arg]
)

_LW_V2_2 = LabwareDefinition2.model_construct(  # type: ignore[call-arg]
    namespace="test",
    version=1,
    schemaVersion=2,
    dimensions=Dimensions(xDimension=1000, yDimension=1200, zDimension=750),
    parameters=Parameters2.model_construct(loadName="labware-name"),  # type: ignore[call-arg]
    gripperOffsets={},
)

_LW_V2_3 = LabwareDefinition2.model_construct(  # type: ignore[call-arg]
    namespace="test",
    version=1,
    schemaVersion=2,
    dimensions=Dimensions(
        xDimension=800,
        yDimension=900,
        zDimension=1000,
    ),
    parameters=Parameters3.model_construct(loadName="unknown-labware-name"),  # type: ignore[call-arg]
    gripperOffsets={},
)

_TC_LID_V2 = LabwareDefinition2.model_construct(  # type: ignore[call-arg]
    namespace="test",
    version=1,
    schemaVersion=2,
    cornerOffsetFromSlot=Vector3D(x=0, y=0, z=0),
    stackingOffsetWithLabware={"default": Vector3D(x=0, y=0, z=0)},
    dimensions=Dimensions(xDimension=1000, yDimension=1200, zDimension=750),
    stackingOffsetWithModule={},
    parameters=Parameters2.model_construct(loadName="tc_lid"),  # type: ignore[call-arg]
    allowedRoles=[LabwareRole.lid],
    gripperOffsets={
        "lidOffsets": GripperOffsets(
            pickUpOffset=Vector3D(x=5, y=10, z=15),
            dropOffset=Vector3D(x=5, y=10, z=15),  # Intentionally same as pickup
        )
    },
)

_AR_LID_V2 = LabwareDefinition2.model_construct(  # type: ignore[call-arg]
    namespace="test",
    version=1,
    schemaVersion=2,
    cornerOffsetFromSlot=Vector3D(x=0, y=0, z=0),
    stackingOffsetWithLabware={"default": Vector3D(x=0, y=0, z=0)},
    dimensions=Dimensions(xDimension=1000, yDimension=1200, zDimension=750),
    stackingOffsetWithModule={},
    parameters=Parameters2.model_construct(
        loadName="opentrons_flex_lid_absorbance_plate_reader_module"
    ),  # type: ignore[call-arg]
    gripperOffsets={
        "default": GripperOffsets(
            pickUpOffset=Vector3D(x=7, y=14, z=21),
            dropOffset=Vector3D(x=28, y=35, z=42),
        )
    },
)

_LW_V3 = LabwareDefinition3.model_construct(  # type: ignore[call-arg]
    namespace="test",
    version=1,
    schemaVersion=3,
    extents=Extents(
        total=AxisAlignedBoundingBox3D(
            backLeftBottom=Vector3D(x=100, y=200, z=300),
            frontRightTop=Vector3D(x=1100, y=-800, z=1300),
        ),
    ),
    features=LocatingFeatures(
        slotFootprintAsChild=SlotFootprintAsChildFeature(
            backLeft=Vector2D(x=-10, y=5), frontRight=Vector2D(x=30, y=-20), z=0
        )
    ),
    parameters=Parameters3.model_construct(loadName="labware-v3-basic"),  # type: ignore[call-arg]
    gripperOffsets={},
)

_LW_V3_WITH_SLOT_FP_AS_CHILD_FEATURE = LabwareDefinition3.model_construct(  # type: ignore[call-arg]
    namespace="test",
    version=1,
    schemaVersion=3,
    extents=Extents(
        total=AxisAlignedBoundingBox3D(
            backLeftBottom=Vector3D(x=50, y=100, z=150),
            frontRightTop=Vector3D(x=850, y=-500, z=950),
        ),
    ),
    features=LocatingFeatures(
        slotFootprintAsChild=SlotFootprintAsChildFeature(
            backLeft=Vector2D(x=0, y=0), frontRight=Vector2D(x=80, y=60), z=5
        )
    ),
    legacyStackingOffsetWithLabware={
        "default": Vector3D(x=0, y=0, z=0),
    },
    parameters=Parameters3.model_construct(loadName="labware-v3-child"),  # type: ignore[call-arg]
    gripperOffsets={},
)

_LW_V3_WITH_SLOT_FP_AS_PARENT_FEATURE = LabwareDefinition3.model_construct(  # type: ignore[call-arg]
    namespace="test",
    version=1,
    schemaVersion=3,
    extents=Extents(
        total=AxisAlignedBoundingBox3D(
            backLeftBottom=Vector3D(x=0, y=0, z=0),
            frontRightTop=Vector3D(x=1000, y=800, z=200),
        ),
    ),
    features=LocatingFeatures(
        slotFootprintAsParent=SlotFootprintAsParentFeature(
            backLeft=Vector2D(x=0, y=0), frontRight=Vector2D(x=120, y=90), z=10
        )
    ),
    parameters=Parameters3.model_construct(loadName="parent-labware-v3"),  # type: ignore[call-arg]
    gripperOffsets={},
)

_LW_V3_WITH_SLOT_AS_PARENT_CHILD_FEATURES = LabwareDefinition3.model_construct(  # type: ignore[call-arg]
    namespace="test",
    version=1,
    schemaVersion=3,
    extents=Extents(
        total=AxisAlignedBoundingBox3D(
            backLeftBottom=Vector3D(x=20, y=30, z=40),
            frontRightTop=Vector3D(x=820, y=-470, z=840),
        ),
    ),
    features=LocatingFeatures(
        slotFootprintAsChild=SlotFootprintAsChildFeature(
            backLeft=Vector2D(x=10, y=15), frontRight=Vector2D(x=70, y=45), z=8
        ),
        slotFootprintAsParent=SlotFootprintAsParentFeature(
            backLeft=Vector2D(x=0, y=0), frontRight=Vector2D(x=100, y=80), z=12
        ),
    ),
    parameters=Parameters3.model_construct(loadName="dual-feature-labware"),  # type: ignore[call-arg]
    gripperOffsets={},
)

_LW_V3_WITH_FLEX_TIP_RACK_LID_AS_PARENT = LabwareDefinition3.model_construct(  # type: ignore[call-arg]
    namespace="test",
    version=1,
    schemaVersion=3,
    extents=Extents(
        total=AxisAlignedBoundingBox3D(
            backLeftBottom=Vector3D(x=0, y=0, z=0),
            frontRightTop=Vector3D(x=1000, y=800, z=50),
        ),
    ),
    features=LocatingFeatures(
        opentronsFlexTipRackLidAsParent=OpentronsFlexTipRackLidAsParentFeature(
            matingZ=25
        )
    ),
    parameters=Parameters3.model_construct(loadName="flex-tip-rack-lid-parent"),  # type: ignore[call-arg]
    gripperOffsets={},
)

_LW_V3_WITH_FLEX_TIP_RACK_LID_AS_CHILD = LabwareDefinition3.model_construct(  # type: ignore[call-arg]
    namespace="test",
    version=1,
    schemaVersion=3,
    extents=Extents(
        total=AxisAlignedBoundingBox3D(
            backLeftBottom=Vector3D(x=10, y=20, z=30),
            frontRightTop=Vector3D(x=810, y=-580, z=830),
        ),
    ),
    features=LocatingFeatures(
        opentronsFlexTipRackLidAsChild=OpentronsFlexTipRackLidAsChildFeature(matingZ=15)
    ),
    stackingOffsetWithLabware={
        "default": Vector3D(x=0, y=0, z=0),
    },
    parameters=Parameters3.model_construct(loadName="flex-tip-rack-lid-child"),  # type: ignore[call-arg]
    gripperOffsets={},
)

_LW_V3_WITH_HS_FLAT_ADAPTER = LabwareDefinition3.model_construct(  # type: ignore[call-arg]
    namespace="test",
    version=1,
    schemaVersion=3,
    extents=Extents(
        total=AxisAlignedBoundingBox3D(
            backLeftBottom=Vector3D(x=0, y=0, z=0),
            frontRightTop=Vector3D(x=1000, y=800, z=100),
        ),
    ),
    features=LocatingFeatures(
        heaterShakerUniversalFlatAdapter=HeaterShakerUniversalFlatAdapterFeature(
            flatSupportThermalCouplingZ=50,
            deckLeft={"wallX": 10, "screwCenter": {"x": 100, "y": 300, "z": 10}},
            deckRight={"wallX": -10, "screwCenter": {"x": 500, "y": 700, "z": 50}},
        )
    ),
    parameters=Parameters3.model_construct(loadName="hs-flat-adapter-parent"),  # type: ignore[call-arg]
    gripperOffsets={},
)

_LW_V3_WITH_FLAT_WELL_SUPPORT = LabwareDefinition3.model_construct(  # type: ignore[call-arg]
    namespace="test",
    version=1,
    schemaVersion=3,
    extents=Extents(
        total=AxisAlignedBoundingBox3D(
            backLeftBottom=Vector3D(x=0, y=0, z=0),
            frontRightTop=Vector3D(x=800, y=600, z=200),
        ),
    ),
    features=LocatingFeatures(
        flatSupportThermalCouplingAsChild=FlatSupportThermalCouplingAsChildFeature(
            wellExteriorBottomZ=25
        )
    ),
    stackingOffsetWithLabware={
        "default": Vector3D(x=0, y=0, z=0),
    },
    parameters=Parameters3.model_construct(loadName="flat-well-support-child"),  # type: ignore[call-arg]
    gripperOffsets={},
)

_LW_V3_WITH_SCREW_ANCHORED_AS_PARENT = LabwareDefinition3.model_construct(  # type: ignore[call-arg]
    namespace="test",
    version=1,
    schemaVersion=3,
    extents=Extents(
        total=AxisAlignedBoundingBox3D(
            backLeftBottom=Vector3D(x=0, y=0, z=0),
            frontRightTop=Vector3D(x=1000, y=800, z=100),
        ),
    ),
    features=LocatingFeatures(
        screwAnchoredAsParent=ScrewAnchoredAsParentFeature(
            screwCenter={"x": 500, "y": 400, "z": 20}
        )
    ),
    parameters=Parameters3.model_construct(loadName="screw-anchored-parent"),  # type: ignore[call-arg]
    gripperOffsets={},
)

_LW_V3_WITH_SCREW_ANCHORED_AS_CHILD = LabwareDefinition3.model_construct(  # type: ignore[call-arg]
    namespace="test",
    version=1,
    schemaVersion=3,
    extents=Extents(
        total=AxisAlignedBoundingBox3D(
            backLeftBottom=Vector3D(x=0, y=0, z=0),
            frontRightTop=Vector3D(x=800, y=600, z=200),
        ),
    ),
    features=LocatingFeatures(
        screwAnchoredAsChild=ScrewAnchoredAsChildFeature(
            screwCenter={"x": 400, "y": 300, "z": 10}
        )
    ),
    stackingOffsetWithLabware={
        "default": Vector3D(x=0, y=0, z=0),
    },
    parameters=Parameters3.model_construct(loadName="screw-anchored-child"),  # type: ignore[call-arg]
    gripperOffsets={},
)

_LW_V3_WITH_LEGACY_STACKING = LabwareDefinition3.model_construct(  # type: ignore[call-arg]
    namespace="test",
    version=1,
    schemaVersion=3,
    extents=Extents(
        total=AxisAlignedBoundingBox3D(
            backLeftBottom=Vector3D(x=75, y=125, z=175),
            frontRightTop=Vector3D(x=875, y=-475, z=975),
        ),
    ),
    features=LocatingFeatures(
        slotFootprintAsChild=SlotFootprintAsChildFeature(
            backLeft=Vector2D(x=5, y=10), frontRight=Vector2D(x=85, y=50), z=3
        )
    ),
    legacyStackingOffsetWithLabware={
        "labware-name": Vector3D(x=25, y=35, z=45),
        "default": Vector3D(x=125, y=225, z=325),
    },
    parameters=Parameters3.model_construct(loadName="labware-v3-with-legacy-stacking"),  # type: ignore[call-arg]
    gripperOffsets={},
)

_LW_V3_WITH_GRIPPER_OFFSETS = LabwareDefinition3.model_construct(  # type: ignore[call-arg]
    namespace="test",
    version=1,
    schemaVersion=3,
    extents=Extents(
        total=AxisAlignedBoundingBox3D(
            backLeftBottom=Vector3D(x=75, y=125, z=175),
            frontRightTop=Vector3D(x=875, y=-475, z=975),
        ),
    ),
    features=LocatingFeatures(
        slotFootprintAsChild=SlotFootprintAsChildFeature(
            backLeft=Vector2D(x=5, y=10), frontRight=Vector2D(x=85, y=50), z=3
        )
    ),
    legacyStackingOffsetWithLabware={
        "default": Vector3D(x=125, y=225, z=325),
    },
    gripperOffsets={
        "default": GripperOffsets(
            pickUpOffset=Vector3D(x=15, y=25, z=35),
            dropOffset=Vector3D(x=45, y=55, z=65),
        )
    },
    parameters=Parameters3.model_construct(loadName="labware-v3-with-gripper"),  # type: ignore[call-arg]
)

# Module definitions
_MODULE_DEF_TEMP_V2 = ModuleDefinition.model_construct(  # type: ignore[call-arg]
    schemaVersion=2,
    model=ModuleModel.TEMPERATURE_MODULE_V2,
    dimensions=ModuleDimensions(
        bareOverallHeight=500,
        overLabwareHeight=600,
        labwareInterfaceXDimension=1000,
        labwareInterfaceYDimension=700,
    ),
    gripperOffsets={},
    orientation={
        "3": "right",
        "6": "right",
        "9": "right",
        "A1": "left",
        "A3": "right",
        "B1": "left",
        "B3": "right",
        "C1": "left",
        "C3": "right",
        "D1": "left",
        "D3": "right",
    },
)

_MODULE_DEF_HS = ModuleDefinition.model_construct(  # type: ignore[call-arg]
    schemaVersion=2,
    model=ModuleModel.HEATER_SHAKER_MODULE_V1,
    dimensions=ModuleDimensions(
        bareOverallHeight=500,
        overLabwareHeight=600,
        labwareInterfaceXDimension=1000,
        labwareInterfaceYDimension=700,
    ),
    gripperOffsets={},
    extents=Extents(
        total=AxisAlignedBoundingBox3D(
            backLeftBottom=Vector3D(x=-18, y=1.8, z=0),
            frontRightTop=Vector3D(x=138.25, y=-89.95, z=82),
        )
    ),
    orientation={
        "3": "right",
        "6": "right",
        "9": "right",
        "A1": "left",
        "A3": "right",
        "B1": "left",
        "B3": "right",
        "C1": "left",
        "C3": "right",
        "D1": "left",
        "D3": "right",
    },
    features={"screwAnchoredAsParent": {"screwCenter": {"x": 64, "y": 43, "z": 0}}},
)

_MODULE_DEF_WITH_GRIPPER_OFFSETS = ModuleDefinition.model_construct(  # type: ignore[call-arg]
    schemaVersion=2,
    model=ModuleModel.HEATER_SHAKER_MODULE_V1,
    dimensions=ModuleDimensions(
        bareOverallHeight=500,
        overLabwareHeight=600,
        labwareInterfaceXDimension=1000,
        labwareInterfaceYDimension=700,
    ),
    gripperOffsets={
        "default": LabwareMovementOffsetData(
            pickUpOffset=LabwareOffsetVector(x=11, y=22, z=33),
            dropOffset=LabwareOffsetVector(x=33, y=22, z=11),
        )
    },
    orientation={
        "3": "right",
        "6": "right",
        "9": "right",
        "A1": "left",
        "A3": "right",
        "B1": "left",
        "B3": "right",
        "C1": "left",
        "C3": "right",
        "D1": "left",
        "D3": "right",
    },
)

_MODULE_DEF_TC_V2 = ModuleDefinition.model_construct(  # type: ignore[call-arg]
    schemaVersion=2,
    model=ModuleModel.THERMOCYCLER_MODULE_V2,
    dimensions=ModuleDimensions(
        bareOverallHeight=1000,
        overLabwareHeight=1100,
    ),
    gripperOffsets={},
    orientation={"B1": "left"},
)

_ADDRESSABLE_AREA = AddressableArea(
    area_name="test_area",
    area_type=AreaType.SLOT,
    base_slot=DeckSlotName.SLOT_A1,
    display_name="Test Area",
    bounding_box=AddressableAreaDimensions(x=1000, y=1500, z=2000),
    position=AddressableOffsetVector(x=0, y=0, z=0),
    compatible_module_types=[],
    features=LocatingFeatures(),
    mating_surface_unit_vector=[-1, 1, -1],
    orientation=ModuleOrientation.NOT_APPLICABLE,
)

_ADDRESSABLE_AREA_WITH_PARENT_FEATURES = AddressableArea(
    area_name="test_area_with_parent",
    area_type=AreaType.SLOT,
    base_slot=DeckSlotName.SLOT_A2,
    display_name="Test Area with Parent Features",
    bounding_box=AddressableAreaDimensions(x=1200, y=1600, z=2200),
    position=AddressableOffsetVector(x=100, y=200, z=300),
    compatible_module_types=[],
    features=LocatingFeatures(
        slotFootprintAsParent=SlotFootprintAsParentFeature(
            backLeft=Vector2D(x=0, y=0), frontRight=Vector2D(x=150, y=120), z=15
        )
    ),
    mating_surface_unit_vector=[-1, 1, -1],
    orientation=ModuleOrientation.NOT_APPLICABLE,
)

_ADDRESSABLE_AREA_WITH_FLEX_TIP_RACK_LID = AddressableArea(
    area_name="test_area_with_flex_lid",
    area_type=AreaType.SLOT,
    base_slot=DeckSlotName.SLOT_A3,
    display_name="Test Area with Flex Tip Rack Lid",
    bounding_box=AddressableAreaDimensions(x=1100, y=1400, z=2100),
    position=AddressableOffsetVector(x=50, y=100, z=150),
    compatible_module_types=[],
    features=LocatingFeatures(
        opentronsFlexTipRackLidAsParent=OpentronsFlexTipRackLidAsParentFeature(
            matingZ=30
        )
    ),
    mating_surface_unit_vector=[-1, 1, -1],
    orientation=ModuleOrientation.NOT_APPLICABLE,
)


class ModuleOverlapSpec(NamedTuple):
    """Spec data to test module overlap behavior."""

    spec_deck_definition: DeckDefinitionV5
    module_definition: ModuleDefinition
    child_definition: LabwareDefinition2
    module_parent_to_child_offset: Point
    labware_location: ModuleLocation
    expected_total_offset: Point


class LabwareOverlapSpec(NamedTuple):
    """Spec data to test labware stacking behavior."""

    child_definition: LabwareDefinition
    parent_definition: LabwareDefinition
    labware_location: OnLabwareLocation
    expected_total_offset: Point


class AddressableAreaSpec(NamedTuple):
    """Spec data to test addressable area behavior."""

    child_definition: LabwareDefinition2
    addressable_area: AddressableArea
    labware_location: AddressableAreaLocation
    expected_total_offset: Point


class LabwareV3Spec(NamedTuple):
    """Spec data to test LabwareDefinition3 behavior."""

    child_definition: LabwareDefinition3
    parent_definition: object
    labware_location: object
    slot_name: DeckSlotName
    expected_total_offset: Point
    underlying_ancestor_definition: LabwareStackupAncestorDefinition


class GripperOffsetSpec(NamedTuple):
    """Spec data to test gripper offset behavior."""

    stackup_lw_info_top_to_bottom: List[tuple[LabwareDefinition, object]]
    underlying_ancestor_definition: object
    slot_name: DeckSlotName
    deck_definition: DeckDefinitionV5
    module_parent_to_child_offset: Optional[Point]
    expected_pick_up_offset: Point
    expected_drop_offset: Point


MODULE_OVERLAP_SPECS: List[ModuleOverlapSpec] = [
    ModuleOverlapSpec(
        spec_deck_definition=load_deck(STANDARD_OT2_DECK, 5),
        module_definition=_MODULE_DEF_TEMP_V2,
        child_definition=_LW_V2_WITH_MODULE_STACKING,
        module_parent_to_child_offset=Point(x=450, y=550, z=650),
        labware_location=ModuleLocation(moduleId="module-1"),
        expected_total_offset=Point(x=550, y=700, z=850),
    ),
    ModuleOverlapSpec(
        spec_deck_definition=load_deck(STANDARD_OT2_DECK, 5),
        module_definition=_MODULE_DEF_TC_V2,
        child_definition=_LW_V2,
        module_parent_to_child_offset=Point(x=450, y=550, z=650),
        labware_location=ModuleLocation(moduleId="module-1"),
        expected_total_offset=Point(x=600, y=800, z=989.3),
    ),
    ModuleOverlapSpec(
        spec_deck_definition=load_deck(STANDARD_OT3_DECK, 5),
        module_definition=_MODULE_DEF_TC_V2,
        child_definition=_LW_V2,
        module_parent_to_child_offset=Point(x=450, y=550, z=650),
        labware_location=ModuleLocation(moduleId="module-1"),
        expected_total_offset=Point(x=600, y=800, z=1000),
    ),
    ModuleOverlapSpec(
        spec_deck_definition=load_deck(STANDARD_OT3_DECK, 5),
        module_definition=_MODULE_DEF_TC_V2,
        child_definition=_LW_V2_WITH_MODULE_STACKING,
        module_parent_to_child_offset=Point(x=450, y=550, z=650),
        labware_location=ModuleLocation(moduleId="module-1"),
        expected_total_offset=Point(x=100, y=200, z=300),
    ),
]

LABWARE_OVERLAP_SPECS: List[LabwareOverlapSpec] = [
    LabwareOverlapSpec(
        child_definition=_LW_V2_WITH_LABWARE_STACKING,
        parent_definition=_LW_V2_2,
        labware_location=OnLabwareLocation(labwareId="parent-labware-1"),
        expected_total_offset=Point(x=250, y=400, z=1000),
    ),
    LabwareOverlapSpec(
        child_definition=_LW_V2_WITH_LABWARE_STACKING,
        parent_definition=_LW_V2_3,
        labware_location=OnLabwareLocation(labwareId="parent-labware-2"),
        expected_total_offset=Point(x=450, y=650, z=950),
    ),
    LabwareOverlapSpec(
        child_definition=_LW_V2_WITH_LABWARE_STACKING,
        parent_definition=_LW_V3_WITH_SLOT_AS_PARENT_CHILD_FEATURES,
        labware_location=OnLabwareLocation(labwareId="parent-labware-3"),
        expected_total_offset=Point(x=440, y=2135, z=742),
    ),
    LabwareOverlapSpec(
        child_definition=_LW_V3_WITH_LEGACY_STACKING,
        parent_definition=_LW_V2_2,
        labware_location=OnLabwareLocation(labwareId="parent-v2-labware"),
        expected_total_offset=Point(x=100, y=385, z=880),
    ),
]

ADDRESSABLE_AREA_SPECS: List[AddressableAreaSpec] = [
    AddressableAreaSpec(
        child_definition=_LW_V2,
        addressable_area=_ADDRESSABLE_AREA,
        labware_location=AddressableAreaLocation(addressableAreaName="test_area"),
        expected_total_offset=Point(x=150, y=250, z=350),
    ),
]

LW_V3_SPECS_ON_NON_LW: List[LabwareV3Spec] = [
    LabwareV3Spec(
        child_definition=_LW_V3,
        parent_definition=_ADDRESSABLE_AREA,
        slot_name=DeckSlotName.SLOT_A1,
        labware_location=AddressableAreaLocation(addressableAreaName="test_area"),
        expected_total_offset=Point(x=10, y=1495, z=0),
        underlying_ancestor_definition=_ADDRESSABLE_AREA,
    ),
    LabwareV3Spec(
        child_definition=_LW_V3_WITH_SLOT_FP_AS_CHILD_FEATURE,
        parent_definition=_ADDRESSABLE_AREA_WITH_PARENT_FEATURES,
        slot_name=DeckSlotName.SLOT_A1,
        labware_location=AddressableAreaLocation(
            addressableAreaName="test_area_with_parent"
        ),
        expected_total_offset=Point(x=0, y=1600, z=-5),
        underlying_ancestor_definition=_ADDRESSABLE_AREA,
    ),
    LabwareV3Spec(
        child_definition=_LW_V3_WITH_FLEX_TIP_RACK_LID_AS_CHILD,
        parent_definition=_ADDRESSABLE_AREA_WITH_FLEX_TIP_RACK_LID,
        slot_name=DeckSlotName.SLOT_A1,
        labware_location=AddressableAreaLocation(
            addressableAreaName="test_area_with_flex_lid"
        ),
        expected_total_offset=Point(x=0, y=0, z=45),
        underlying_ancestor_definition=_ADDRESSABLE_AREA,
    ),
]

LW_V3_SPECS_ON_LW: List[LabwareV3Spec] = [
    LabwareV3Spec(
        child_definition=_LW_V3_WITH_SLOT_FP_AS_CHILD_FEATURE,
        parent_definition=_LW_V3_WITH_SLOT_FP_AS_PARENT_FEATURE,
        slot_name=DeckSlotName.SLOT_A1,
        labware_location=OnLabwareLocation(labwareId="parent-labware-v3"),
        expected_total_offset=Point(x=20.0, y=15, z=5),
        underlying_ancestor_definition=_ADDRESSABLE_AREA,
    ),
    LabwareV3Spec(
        child_definition=_LW_V3_WITH_SLOT_FP_AS_CHILD_FEATURE,
        parent_definition=_LW_V3,
        slot_name=DeckSlotName.SLOT_A1,
        labware_location=OnLabwareLocation(labwareId="labware-v3-basic"),
        expected_total_offset=Point(x=10, y=1495, z=1000),
        underlying_ancestor_definition=_ADDRESSABLE_AREA,
    ),
    LabwareV3Spec(
        child_definition=_LW_V3_WITH_FLEX_TIP_RACK_LID_AS_CHILD,
        parent_definition=_LW_V3_WITH_FLEX_TIP_RACK_LID_AS_PARENT,
        slot_name=DeckSlotName.SLOT_A1,
        labware_location=OnLabwareLocation(labwareId="flex-tip-rack-lid-parent"),
        expected_total_offset=Point(x=0, y=0, z=40),
        underlying_ancestor_definition=_ADDRESSABLE_AREA,
    ),
    LabwareV3Spec(
        child_definition=_LW_V3_WITH_FLAT_WELL_SUPPORT,
        parent_definition=_LW_V3_WITH_HS_FLAT_ADAPTER,
        labware_location=OnLabwareLocation(labwareId="hs-flat-adapter-parent"),
        slot_name=DeckSlotName.SLOT_D1,
        expected_total_offset=Point(x=-26, y=-157.0, z=25),
        underlying_ancestor_definition=_MODULE_DEF_HS,
    ),
    LabwareV3Spec(
        child_definition=_LW_V3_WITH_FLAT_WELL_SUPPORT,
        parent_definition=_LW_V3_WITH_HS_FLAT_ADAPTER,
        labware_location=OnLabwareLocation(labwareId="hs-flat-adapter-parent"),
        slot_name=DeckSlotName.SLOT_D3,
        expected_total_offset=Point(x=-246, y=-557.0, z=25),
        underlying_ancestor_definition=_MODULE_DEF_HS,
    ),
    LabwareV3Spec(
        child_definition=_LW_V3_WITH_SCREW_ANCHORED_AS_CHILD,
        parent_definition=_LW_V3_WITH_SCREW_ANCHORED_AS_PARENT,
        labware_location=OnLabwareLocation(labwareId="screw-anchored-parent"),
        slot_name=DeckSlotName.SLOT_A1,
        expected_total_offset=Point(x=100, y=100, z=10),
        underlying_ancestor_definition=_ADDRESSABLE_AREA,
    ),
    LabwareV3Spec(
        child_definition=_LW_V3_WITH_HS_FLAT_ADAPTER,
        parent_definition=_LW_V3_WITH_SCREW_ANCHORED_AS_PARENT,
        labware_location=OnLabwareLocation(labwareId="screw-anchored-parent"),
        slot_name=DeckSlotName.SLOT_D1,
        expected_total_offset=Point(x=400, y=100, z=20),
        underlying_ancestor_definition=_MODULE_DEF_HS,
    ),
    LabwareV3Spec(
        child_definition=_LW_V3_WITH_HS_FLAT_ADAPTER,
        parent_definition=_LW_V3_WITH_SCREW_ANCHORED_AS_PARENT,
        labware_location=OnLabwareLocation(labwareId="screw-anchored-parent"),
        slot_name=DeckSlotName.SLOT_D3,
        expected_total_offset=Point(x=0, y=-300, z=20),
        underlying_ancestor_definition=_MODULE_DEF_HS,
    ),
]


GRIPPER_OFFSET_SPECS: List[GripperOffsetSpec] = [
    GripperOffsetSpec(
        stackup_lw_info_top_to_bottom=[
            (_LW_V2, DeckSlotLocation(slotName=DeckSlotName.SLOT_3)),
        ],
        underlying_ancestor_definition=_ADDRESSABLE_AREA,
        slot_name=DeckSlotName.SLOT_3,
        deck_definition=load_deck(STANDARD_OT3_DECK, 5),
        module_parent_to_child_offset=None,
        expected_pick_up_offset=Point(x=0, y=0, z=0),
        expected_drop_offset=Point(x=0, y=0, z=-0.75),
    ),
    GripperOffsetSpec(
        stackup_lw_info_top_to_bottom=[
            (_LW_V2, ModuleLocation(moduleId="module-id")),
        ],
        underlying_ancestor_definition=_MODULE_DEF_WITH_GRIPPER_OFFSETS,
        slot_name=DeckSlotName.SLOT_A1,
        deck_definition=load_deck(STANDARD_OT3_DECK, 5),
        module_parent_to_child_offset=Point(x=0, y=0, z=0),
        expected_pick_up_offset=Point(x=11, y=22, z=33),
        expected_drop_offset=Point(x=33, y=22, z=11),
    ),
    GripperOffsetSpec(
        stackup_lw_info_top_to_bottom=[
            (
                _LW_V2_WITH_DEFAULT_GRIPPER_OFFSETS,
                OnLabwareLocation(labwareId="parent-id"),
            ),
            (_LW_V2, DeckSlotLocation(slotName=DeckSlotName.SLOT_1)),
        ],
        underlying_ancestor_definition=_ADDRESSABLE_AREA,
        slot_name=DeckSlotName.SLOT_1,
        deck_definition=load_deck(STANDARD_OT3_DECK, 5),
        module_parent_to_child_offset=None,
        expected_pick_up_offset=Point(x=10, y=20, z=30),
        expected_drop_offset=Point(x=40, y=50, z=59.25),
    ),
    GripperOffsetSpec(
        stackup_lw_info_top_to_bottom=[
            (_LW_V2, OnLabwareLocation(labwareId="adapter-id")),
            (
                _LW_V2_WITH_SLOT_SPECIFIC_GRIPPER_OFFSETS,
                ModuleLocation(moduleId="module-id"),
            ),
        ],
        underlying_ancestor_definition=_MODULE_DEF_WITH_GRIPPER_OFFSETS,
        slot_name=DeckSlotName.SLOT_C1,
        deck_definition=load_deck(STANDARD_OT3_DECK, 5),
        module_parent_to_child_offset=Point(x=0, y=0, z=0),
        expected_pick_up_offset=Point(x=111, y=222, z=333),
        expected_drop_offset=Point(x=333, y=222, z=111),
    ),
    GripperOffsetSpec(
        stackup_lw_info_top_to_bottom=[
            (_TC_LID_V2, OnLabwareLocation(labwareId="adapter-id")),
            (_LW_V2, ModuleLocation(moduleId="module-id")),
        ],
        underlying_ancestor_definition=_MODULE_DEF_TC_V2,
        slot_name=DeckSlotName.SLOT_B1,
        deck_definition=load_deck(STANDARD_OT3_DECK, 5),
        module_parent_to_child_offset=Point(x=0, y=0, z=0),
        expected_pick_up_offset=Point(x=5, y=10, z=15),
        expected_drop_offset=Point(x=5, y=10, z=15),
    ),
    GripperOffsetSpec(
        stackup_lw_info_top_to_bottom=[
            (_AR_LID_V2, DeckSlotLocation(slotName=DeckSlotName.SLOT_3)),
        ],
        underlying_ancestor_definition=_ADDRESSABLE_AREA,
        slot_name=DeckSlotName.SLOT_3,
        deck_definition=load_deck(STANDARD_OT3_DECK, 5),
        module_parent_to_child_offset=None,
        expected_pick_up_offset=Point(x=7, y=14, z=21),
        expected_drop_offset=Point(x=28, y=35, z=41.25),
    ),
    GripperOffsetSpec(
        stackup_lw_info_top_to_bottom=[
            (_LW_V3_WITH_GRIPPER_OFFSETS, OnLabwareLocation(labwareId="parent-id")),
            (
                _LW_V2_WITH_DEFAULT_GRIPPER_OFFSETS,
                DeckSlotLocation(slotName=DeckSlotName.SLOT_1),
            ),
        ],
        underlying_ancestor_definition=_ADDRESSABLE_AREA,
        slot_name=DeckSlotName.SLOT_1,
        deck_definition=load_deck(STANDARD_OT3_DECK, 5),
        module_parent_to_child_offset=None,
        expected_pick_up_offset=Point(x=15.0, y=25.0, z=35.0),
        expected_drop_offset=Point(x=45.0, y=55.0, z=64.25),
    ),
]


@pytest.mark.parametrize(
    argnames=ModuleOverlapSpec._fields,
    argvalues=MODULE_OVERLAP_SPECS,
)
def test_get_parent_placement_origin_to_lw_origin_with_module(
    spec_deck_definition: DeckDefinitionV5,
    module_definition: ModuleDefinition,
    child_definition: LabwareDefinition2,
    module_parent_to_child_offset: Point,
    labware_location: ModuleLocation,
    expected_total_offset: Point,
) -> None:
    """It should calculate the correct offset from module parent to labware origin."""
    result = get_stackup_origin_to_labware_origin(
        context=LabwareOriginContext.PIPETTING,
        stackup_lw_info_top_to_bottom=[
            (child_definition, labware_location),
        ],
        underlying_ancestor_definition=module_definition,
        slot_name=DeckSlotName.SLOT_B1,
        module_parent_to_child_offset=module_parent_to_child_offset,
        deck_definition=spec_deck_definition,
    )

    assert result == expected_total_offset


@pytest.mark.parametrize(
    argnames=LabwareOverlapSpec._fields,
    argvalues=LABWARE_OVERLAP_SPECS,
)
def test_get_parent_placement_origin_to_lw_origin_with_labware(
    child_definition: LabwareDefinition2,
    parent_definition: LabwareDefinition2,
    labware_location: OnLabwareLocation,
    expected_total_offset: Point,
) -> None:
    """It should calculate the correct offset from labware parent to labware origin for v2_schema_labware."""
    result = get_stackup_origin_to_labware_origin(
        context=LabwareOriginContext.PIPETTING,
        stackup_lw_info_top_to_bottom=[
            (child_definition, labware_location),
            (
                parent_definition,
                AddressableAreaLocation(addressableAreaName="test_area"),
            ),
        ],
        underlying_ancestor_definition=_ADDRESSABLE_AREA,
        slot_name=DeckSlotName.SLOT_A1,
        module_parent_to_child_offset=None,
        deck_definition=load_deck(STANDARD_OT3_DECK, 5),
    )

    assert result == expected_total_offset


@pytest.mark.parametrize(
    argnames=AddressableAreaSpec._fields,
    argvalues=ADDRESSABLE_AREA_SPECS,
)
def test_get_parent_placement_origin_to_lw_origin_with_addressable_area(
    child_definition: LabwareDefinition2,
    addressable_area: AddressableArea,
    labware_location: AddressableAreaLocation,
    expected_total_offset: Point,
) -> None:
    """It should calculate the correct offset from addressable area to labware origin."""
    result = get_stackup_origin_to_labware_origin(
        context=LabwareOriginContext.PIPETTING,
        stackup_lw_info_top_to_bottom=[
            (child_definition, labware_location),
        ],
        underlying_ancestor_definition=addressable_area,
        slot_name=DeckSlotName.SLOT_A1,
        module_parent_to_child_offset=None,
        deck_definition=load_deck(STANDARD_OT3_DECK, 5),
    )

    assert result == expected_total_offset


@pytest.mark.parametrize(
    argnames=LabwareV3Spec._fields,
    argvalues=LW_V3_SPECS_ON_NON_LW,
)
def test_get_parent_placement_origin_to_lw_origin_v3_definitions_non_lw(
    child_definition: LabwareDefinition3,
    parent_definition: AddressableArea,
    labware_location: AddressableAreaLocation,
    underlying_ancestor_definition: LabwareStackupAncestorDefinition,
    slot_name: DeckSlotName,
    expected_total_offset: Point,
) -> None:
    """It should handle LabwareDefinition3 correctly with various parent configurations (that are not labware)."""
    result = get_stackup_origin_to_labware_origin(
        context=LabwareOriginContext.PIPETTING,
        stackup_lw_info_top_to_bottom=[
            (child_definition, labware_location),
        ],
        underlying_ancestor_definition=parent_definition,
        slot_name=slot_name,
        module_parent_to_child_offset=None,
        deck_definition=load_deck(STANDARD_OT3_DECK, 5),
    )

    assert result == expected_total_offset


@pytest.mark.parametrize(
    argnames=LabwareV3Spec._fields,
    argvalues=LW_V3_SPECS_ON_LW,
)
def test_get_parent_placement_origin_to_lw_origin_v3_definitions(
    child_definition: LabwareDefinition3,
    parent_definition: LabwareDefinition3,
    labware_location: OnLabwareLocation,
    expected_total_offset: Point,
    underlying_ancestor_definition: LabwareStackupAncestorDefinition,
    slot_name: DeckSlotName,
) -> None:
    """It should handle LabwareDefinition3 correctly with various labware configurations."""
    result = get_stackup_origin_to_labware_origin(
        context=LabwareOriginContext.PIPETTING,
        stackup_lw_info_top_to_bottom=[
            (child_definition, labware_location),
            (
                parent_definition,
                AddressableAreaLocation(addressableAreaName="test_area"),
            ),
        ],
        underlying_ancestor_definition=underlying_ancestor_definition,
        slot_name=slot_name,
        module_parent_to_child_offset=None,
        deck_definition=load_deck(STANDARD_OT3_DECK, 5),
    )

    assert result == expected_total_offset


def test_v3_child_on_v2_parent_labware() -> None:
    """Test that v3 labware can stack on v2 labware correctly."""
    deck_def = load_deck(STANDARD_OT3_DECK, 5)

    # Top-most labware case
    result = get_stackup_origin_to_labware_origin(
        context=LabwareOriginContext.PIPETTING,
        stackup_lw_info_top_to_bottom=[
            (_LW_V3_WITH_LEGACY_STACKING, OnLabwareLocation(labwareId="parent-id")),
            (_LW_V2_2, AddressableAreaLocation(addressableAreaName="test_area")),
        ],
        underlying_ancestor_definition=_ADDRESSABLE_AREA,
        slot_name=DeckSlotName.SLOT_A1,
        module_parent_to_child_offset=None,
        deck_definition=deck_def,
    )

    assert result == Point(x=100, y=385, z=880)

    # Non-topmost labware case
    result_non_topmost = get_stackup_origin_to_labware_origin(
        context=LabwareOriginContext.PIPETTING,
        stackup_lw_info_top_to_bottom=[
            (_LW_V2, OnLabwareLocation(labwareId="adapter-id")),
            (_LW_V3_WITH_LEGACY_STACKING, OnLabwareLocation(labwareId="parent-id")),
            (_LW_V2_2, AddressableAreaLocation(addressableAreaName="test_area")),
        ],
        underlying_ancestor_definition=_ADDRESSABLE_AREA,
        slot_name=DeckSlotName.SLOT_A1,
        module_parent_to_child_offset=None,
        deck_definition=deck_def,
    )

    expected_total = Point(x=175, y=760, z=1855)

    assert result_non_topmost == expected_total


@pytest.mark.parametrize(
    argnames=GripperOffsetSpec._fields,
    argvalues=GRIPPER_OFFSET_SPECS,
)
def test_gripper_offsets_for_pickup(
    stackup_lw_info_top_to_bottom: List[tuple[LabwareDefinition, LabwareLocation]],
    underlying_ancestor_definition: LabwareStackupAncestorDefinition,
    slot_name: DeckSlotName,
    deck_definition: DeckDefinitionV5,
    module_parent_to_child_offset: Optional[Point],
    expected_pick_up_offset: Point,
    expected_drop_offset: Point,
) -> None:
    """It should calculate the correct gripper offsets for pick up context."""
    pick_up_result = get_stackup_origin_to_labware_origin(
        context=LabwareOriginContext.GRIPPER_PICKING_UP,
        stackup_lw_info_top_to_bottom=stackup_lw_info_top_to_bottom,
        underlying_ancestor_definition=underlying_ancestor_definition,
        slot_name=slot_name,
        module_parent_to_child_offset=module_parent_to_child_offset,
        deck_definition=deck_definition,
    )

    # Baseline without gripper offsets
    pipetting_result = get_stackup_origin_to_labware_origin(
        context=LabwareOriginContext.PIPETTING,
        stackup_lw_info_top_to_bottom=stackup_lw_info_top_to_bottom,
        underlying_ancestor_definition=underlying_ancestor_definition,
        slot_name=slot_name,
        module_parent_to_child_offset=module_parent_to_child_offset,
        deck_definition=deck_definition,
    )

    actual_pick_up_offset = pick_up_result - pipetting_result

    assert actual_pick_up_offset == expected_pick_up_offset


@pytest.mark.parametrize(
    argnames=GripperOffsetSpec._fields,
    argvalues=GRIPPER_OFFSET_SPECS,
)
def test_gripper_offsets_for_drop(
    stackup_lw_info_top_to_bottom: List[tuple[LabwareDefinition, LabwareLocation]],
    underlying_ancestor_definition: LabwareStackupAncestorDefinition,
    slot_name: DeckSlotName,
    deck_definition: DeckDefinitionV5,
    module_parent_to_child_offset: Optional[Point],
    expected_drop_offset: Point,
    expected_pick_up_offset: Point,
) -> None:
    """It should calculate the correct gripper offsets for drop context."""
    drop_result = get_stackup_origin_to_labware_origin(
        context=LabwareOriginContext.GRIPPER_DROPPING,
        stackup_lw_info_top_to_bottom=stackup_lw_info_top_to_bottom,
        underlying_ancestor_definition=underlying_ancestor_definition,
        slot_name=slot_name,
        module_parent_to_child_offset=module_parent_to_child_offset,
        deck_definition=deck_definition,
    )

    # Baseline without gripper offsets
    pipetting_result = get_stackup_origin_to_labware_origin(
        context=LabwareOriginContext.PIPETTING,
        stackup_lw_info_top_to_bottom=stackup_lw_info_top_to_bottom,
        underlying_ancestor_definition=underlying_ancestor_definition,
        slot_name=slot_name,
        module_parent_to_child_offset=module_parent_to_child_offset,
        deck_definition=deck_definition,
    )

    actual_drop_offset = drop_result - pipetting_result

    assert actual_drop_offset == expected_drop_offset


def test_gripper_offsets_stacked_labware_with_module() -> None:
    """Test labware on adapter on module with slot-specific offsets."""
    deck_def = load_deck(STANDARD_OT3_DECK, 5)

    stackup_lw_info: list[tuple[LabwareDefinition, LabwareLocation]] = [
        (_LW_V2, OnLabwareLocation(labwareId="adapter-id")),
        (
            _LW_V2_WITH_SLOT_SPECIFIC_GRIPPER_OFFSETS,
            ModuleLocation(moduleId="module-id"),
        ),
    ]

    pick_up_result = get_stackup_origin_to_labware_origin(
        context=LabwareOriginContext.GRIPPER_PICKING_UP,
        stackup_lw_info_top_to_bottom=stackup_lw_info,
        underlying_ancestor_definition=_MODULE_DEF_WITH_GRIPPER_OFFSETS,
        slot_name=DeckSlotName.SLOT_C1,
        module_parent_to_child_offset=Point(x=0, y=0, z=0),
        deck_definition=deck_def,
    )

    pipetting_result = get_stackup_origin_to_labware_origin(
        context=LabwareOriginContext.PIPETTING,
        stackup_lw_info_top_to_bottom=stackup_lw_info,
        underlying_ancestor_definition=_MODULE_DEF_WITH_GRIPPER_OFFSETS,
        slot_name=DeckSlotName.SLOT_C1,
        module_parent_to_child_offset=Point(x=0, y=0, z=0),
        deck_definition=deck_def,
    )

    expected_pick_up_offset = Point(x=111, y=222, z=333)
    actual_pick_up_offset = pick_up_result - pipetting_result

    assert actual_pick_up_offset == expected_pick_up_offset

    drop_result = get_stackup_origin_to_labware_origin(
        context=LabwareOriginContext.GRIPPER_DROPPING,
        stackup_lw_info_top_to_bottom=stackup_lw_info,
        underlying_ancestor_definition=_MODULE_DEF_WITH_GRIPPER_OFFSETS,
        slot_name=DeckSlotName.SLOT_C1,
        module_parent_to_child_offset=Point(x=0, y=0, z=0),
        deck_definition=deck_def,
    )

    expected_drop_offset = Point(x=333, y=222, z=111)
    actual_drop_offset = drop_result - pipetting_result

    assert actual_drop_offset == expected_drop_offset


def test_gripper_offsets_fallback_to_default() -> None:
    """Test that gripper offsets fall back to default when slot-specific not available."""
    deck_def = load_deck(STANDARD_OT3_DECK, 5)

    stackup_lw_info: list[tuple[LabwareDefinition, LabwareLocation]] = [
        (_LW_V2, OnLabwareLocation(labwareId="adapter-id")),
        (
            _LW_V2_WITH_DEFAULT_GRIPPER_OFFSETS,
            ModuleLocation(moduleId="module-id"),
        ),
    ]

    pick_up_result = get_stackup_origin_to_labware_origin(
        context=LabwareOriginContext.GRIPPER_PICKING_UP,
        stackup_lw_info_top_to_bottom=stackup_lw_info,
        underlying_ancestor_definition=_MODULE_DEF_WITH_GRIPPER_OFFSETS,
        slot_name=DeckSlotName.SLOT_D1,
        module_parent_to_child_offset=Point(x=0, y=0, z=0),
        deck_definition=deck_def,
    )

    # Baseline without gripper offsets
    pipetting_result = get_stackup_origin_to_labware_origin(
        context=LabwareOriginContext.PIPETTING,
        stackup_lw_info_top_to_bottom=stackup_lw_info,
        underlying_ancestor_definition=_MODULE_DEF_WITH_GRIPPER_OFFSETS,
        slot_name=DeckSlotName.SLOT_D1,
        module_parent_to_child_offset=Point(x=0, y=0, z=0),
        deck_definition=deck_def,
    )

    expected_pick_up_offset = Point(x=111, y=222, z=333)
    actual_pick_up_offset = pick_up_result - pipetting_result

    assert actual_pick_up_offset == expected_pick_up_offset


def test_hs_flat_adapter_center_slot_raises_error() -> None:
    """Test that heater shaker flat adapter raises error for center slot."""
    with pytest.raises(
        InvalidLabwarePlacementError,
        match="heaterShakerUniversalFlatAdapter.*center",
    ):
        get_stackup_origin_to_labware_origin(
            context=LabwareOriginContext.PIPETTING,
            stackup_lw_info_top_to_bottom=[
                (
                    _LW_V3_WITH_FLAT_WELL_SUPPORT,
                    OnLabwareLocation(labwareId="hs-flat-adapter-parent"),
                ),
                (
                    _LW_V3_WITH_HS_FLAT_ADAPTER,
                    AddressableAreaLocation(addressableAreaName="test_area"),
                ),
            ],
            underlying_ancestor_definition=_ADDRESSABLE_AREA,
            slot_name=DeckSlotName.SLOT_A2,
            module_parent_to_child_offset=None,
            deck_definition=load_deck(STANDARD_OT3_DECK, 5),
        )



"""Test robot.open-gripper-jaw commands."""

from decoy import Decoy

from opentrons.hardware_control import OT3HardwareControlAPI
from opentrons.protocol_engine.commands.command import SuccessData
from opentrons.protocol_engine.commands.robot.close_gripper_jaw import (
    CloseGripperJawImplementation,
    CloseGripperJawParams,
    CloseGripperJawResult,
)
from opentrons.protocol_engine.state.state import StateView


async def test_close_gripper_jaw_implementation(
    decoy: Decoy,
    state_view: StateView,
    ot3_hardware_api: OT3HardwareControlAPI,
) -> None:
    """Test the `robot.closeGripperJaw` implementation."""
    subject = CloseGripperJawImplementation(
        hardware_api=ot3_hardware_api, state_view=state_view
    )
    decoy.when(state_view.config.use_virtual_gripper).then_return(False)

    params = CloseGripperJawParams(force=10)

    result = await subject.execute(params=params)

    assert result == SuccessData(public=CloseGripperJawResult())
    decoy.verify(await ot3_hardware_api.grip(force_newtons=10))


async def test_close_gripper_jaw_analysis(
    decoy: Decoy,
    state_view: StateView,
    ot3_hardware_api: OT3HardwareControlAPI,
) -> None:
    """Test that closeGripperJaw doesn't call the hardware controller in analysis."""
    subject = CloseGripperJawImplementation(
        hardware_api=ot3_hardware_api, state_view=state_view
    )
    decoy.when(state_view.config.use_virtual_gripper).then_return(True)

    params = CloseGripperJawParams(force=10)

    result = await subject.execute(params=params)

    assert result == SuccessData(public=CloseGripperJawResult())
    decoy.verify(await ot3_hardware_api.grip(force_newtons=10), times=0)



import pytest
from decoy import Decoy

from opentrons.hardware_control.execution_manager import ExecutionManager
from opentrons.hardware_control.modules.types import (
    ModuleDisconnectedCallback,
    ModuleErrorCallback,
)


@pytest.fixture
def module_disconnected_callback(decoy: Decoy) -> ModuleDisconnectedCallback:
    return decoy.mock(cls=ModuleDisconnectedCallback)


@pytest.fixture
def module_error_callback(decoy: Decoy) -> ModuleErrorCallback:
    return decoy.mock(cls=ModuleErrorCallback)


@pytest.fixture
def mock_execution_manager(decoy: Decoy) -> ExecutionManager:
    return decoy.mock(cls=ExecutionManager)



from typing import Iterator, cast

import pytest
from _pytest.fixtures import SubRequest
from mock import MagicMock, patch

from opentrons.drivers.rpi_drivers.types import PortGroup, USBPort
from opentrons.drivers.rpi_drivers.usb import USBBus
from opentrons.hardware_control.modules.types import ModuleAtPort
from opentrons.hardware_control.types import BoardRevision

fake_bus_og = [
    "/sys/bus/usb/devices/usb1/dev",
    "/sys/bus/usb/devices/usb1/1-1/1-1.3/1-1.3:1.0/tty/ttyACM1/dev",
    "/sys/bus/usb/devices/usb1/1-1/1-1.5/1-1.5:1.0/tty/ttyAMA0/dev",
    "/sys/bus/usb/devices/usb1/1-1/1-1.5/1-1.5/1-1.5.1/1-1.5.1:1.0/tty/ttyAMA1/dev",
    "/sys/bus/usb/devices/usb1/1-1/1-1.5/1-1.5/1-1.5.3/1-1.5.3:1.0/tty/ttyACM2/dev",
    "/sys/bus/usb/devices/usb1/1-1/1-1.3/dev",
    "/sys/bus/usb/devices/usb1/1-1/1-1.1/dev",
    "/sys/bus/usb/devices/usb1/1-1/dev",
    "/sys/bus/usb/devices/usb1/1-1/1-1.5/dev",
    "/sys/bus/usb/devices/usb1/1-1/1-1.5/1-1.5:1.2/0003:046D:C52B.0017/hidraw/hidraw0/dev",
    "/sys/bus/usb/devices/usb1/1-1/1-1.5/1-1.5:1.2/usbmisc/hiddev0/dev",
]

fake_bus_refresh = [
    "/sys/bus/usb/devices/usb1/1-1/1-1.3/1-1.3.1/1-1.3.1:1.0/tty/ttyACM1/dev",
    "/sys/bus/usb/devices/usb1/1-1/1-1.3/1-1.3.1/dev",
    "/sys/bus/usb/devices/usb1/1-1/1-1.3/dev",
    "/sys/bus/usb/devices/usb1/1-1/1-1.3/1-1.3.2/dev",
    "/sys/bus/usb/devices/usb1/1-1/1-1.3/1-1.3.2/1-1.3.2.3/1-1.3.2.3:1.0/tty/ttyACM2/dev"
    "/sys/bus/usb/devices/usb1/1-1/1-1.1/dev",
    "/sys/bus/usb/devices/usb1/1-1/dev",
]

fake_bus_flex = [
    "/sys/bus/usb/devices/usb1/dev",
    "/sys/bus/usb/devices/usb1/1-1/1-1.3/dev",
    "/sys/bus/usb/devices/usb1/1-1/1-1.3/1-1.3.4/1-1.3.4:1.0/tty/ttyACM1/dev",
    "/sys/bus/usb/devices/usb1/1-1/1-1.3/1-1.3.4/dev",
    "/sys/bus/usb/devices/usb1/1-1/1-1.1/dev",
    "/sys/bus/usb/devices/usb1/1-1/1-1.1/1-1.1:1.0/tty/ttyACM0/dev",
    "/sys/bus/usb/devices/usb1/1-1/dev",
    "/sys/bus/usb/devices/usb1/1-1/1-1.6/dev",
    "/sys/bus/usb/devices/usb1/1-1/1-1.6/1-1.6:1.0/media0/dev",
    "/sys/bus/usb/devices/usb1/1-1/1-1.6/1-1.6:1.0/video4linux/video1/dev",
    "/sys/bus/usb/devices/usb1/1-1/1-1.6/1-1.6:1.0/video4linux/video0/dev",
    "/sys/bus/usb/devices/usb1/1-1/1-1.6/1-1.6:1.0/input/input3/event3/dev",
    "/sys/bus/usb/devices/usb1/1-1/1-1.4/1-1.4.2/1-1.4.2.4/dev",
    "/sys/bus/usb/devices/usb1/1-1/1-1.4/1-1.4.2/1-1.4.2.4/1-1.4.2.4:1.0/tty/ttyACM3/dev",
    "/sys/bus/usb/devices/usb1/1-1/1-1.4/1-1.4.2/dev",
    "/sys/bus/usb/devices/usb1/1-1/1-1.4/1-1.4.2/1-1.4.2.3/1-1.4.2.3:1.0/tty/ttyACM2/dev",
    "/sys/bus/usb/devices/usb1/1-1/1-1.4/1-1.4.2/1-1.4.2.3/dev",
    "/sys/bus/usb/devices/usb1/1-1/1-1.4/dev",
    "/sys/bus/usb/devices/usb1/1-1/1-1.7/dev",
    "/sys/bus/usb/devices/usb1/1-1/1-1.7/1-1.7:1.0/tty/ttyACM4/dev",
    "/sys/bus/usb/devices/usb1/1-1/1-1.7/1-1.7.4/dev",
    "/sys/bus/usb/devices/usb1/1-1/1-1.7/1-1.7.4/1-1.7.4:1.0/tty/ttyACM5/dev",
]


@pytest.fixture(params=[BoardRevision.OG, BoardRevision.A, BoardRevision.FLEX_B2])
def revision(request: SubRequest) -> BoardRevision:
    return cast(BoardRevision, request.param)


@pytest.fixture()
def usb_bus(revision: BoardRevision) -> Iterator[USBBus]:
    @staticmethod  # type: ignore[misc]
    def fake_read_bus() -> list[str]:
        if revision == BoardRevision.OG:
            return fake_bus_og
        elif revision == BoardRevision.A:
            return fake_bus_refresh
        else:
            return fake_bus_flex

    # TODO(mc, 2022-03-01): partial patching the class under test creates
    # a contaminated test subject that reduces the value of these tests
    # https://github.com/testdouble/contributing-tests/wiki/Partial-Mock
    with patch.object(USBBus, "_read_bus", fake_read_bus):
        yield USBBus(revision)


def test_modify_module_list(revision: BoardRevision, usb_bus: USBBus) -> None:
    # TODO(mc, 2022-03-01): partial patching the class under test creates
    # a contaminated test subject that reduces the value of these tests
    # https://github.com/testdouble/contributing-tests/wiki/Partial-Mock
    usb_bus._read_symlink = MagicMock(return_value="ttyACM1")  # type: ignore[method-assign]
    mod_at_port_list = [
        ModuleAtPort(
            name="temperature module", port="dev/ot_module_temperature_module"
        ),
    ]
    updated_list = usb_bus.match_virtual_ports(mod_at_port_list)

    if revision == BoardRevision.OG:
        expected_name = "1-1.3"
        assert updated_list[0].usb_port == USBPort(
            name=expected_name,
            port_number=1,
            port_group=PortGroup.MAIN,
            device_path="1.0/tty/ttyACM1/dev",
            hub=False,
            hub_port=None,
        )
    elif revision == BoardRevision.A:
        expected_name = "1-1.3.1"
        assert updated_list[0].usb_port == USBPort(
            name=expected_name,
            port_number=1,
            port_group=PortGroup.MAIN,
            device_path="1.0/tty/ttyACM1/dev",
            hub=False,
            hub_port=None,
        )
    elif revision == BoardRevision.FLEX_B2:
        expected_name = "1-1.3.4"
        assert updated_list[0].usb_port == USBPort(
            name=expected_name,
            port_number=5,
            port_group=PortGroup.RIGHT,
            device_path="1.0/tty/ttyACM1/dev",
            hub=False,
            hub_port=None,
        )

    usb_bus._read_symlink = MagicMock(return_value="ttyACM2")  # type: ignore[method-assign]
    mod_at_port_list = [
        ModuleAtPort(name="magnetic module", port="dev/ot_module_magnetic_module"),
    ]
    updated_list = usb_bus.match_virtual_ports(mod_at_port_list)

    if revision == BoardRevision.OG:
        expected_name = "1-1.5.3"
        assert updated_list[0].usb_port == USBPort(
            name=expected_name,
            port_number=2,
            port_group=PortGroup.MAIN,
            device_path="1.0/tty/ttyACM2/dev",
            hub=True,
            hub_port=3,
        )
    elif revision == BoardRevision.A:
        expected_name = "1-1.3.2.3"
        assert updated_list[0].usb_port == USBPort(
            name=expected_name,
            port_number=2,
            port_group=PortGroup.MAIN,
            device_path="1.0/tty/ttyACM2/dev",
            hub=True,
            hub_port=3,
        )
    elif revision == BoardRevision.FLEX_B2:
        expected_name = "1-1.4.2.3"
        assert updated_list[0].usb_port == USBPort(
            name=expected_name,
            port_number=3,
            port_group=PortGroup.LEFT,
            device_path="1.0/tty/ttyACM2/dev",
            hub=True,
            hub_port=3,
        )

    if revision == BoardRevision.FLEX_B2:
        usb_bus._read_symlink = MagicMock(return_value="ttyACM4")  # type: ignore[method-assign]
        mod_at_port_list = [
            ModuleAtPort(
                name="heater-shaker module", port="dev/ot_module_heater_shaker_module"
            ),
        ]
        updated_list = usb_bus.match_virtual_ports(mod_at_port_list)
        expected_name = "1-1.7"
        assert updated_list[0].usb_port == USBPort(
            name=expected_name,
            port_number=9,
            port_group=PortGroup.FRONT,
            device_path="1.0/tty/ttyACM4/dev",
            hub=False,
            hub_port=None,
        )

        usb_bus._read_symlink = MagicMock(return_value="ttyACM5")  # type: ignore[method-assign]
        mod_at_port_list = [
            ModuleAtPort(
                name="thermocycler module", port="dev/ot_module_thermocycler_module"
            ),
        ]
        updated_list = usb_bus.match_virtual_ports(mod_at_port_list)
        expected_name = "1-1.7.4"
        assert updated_list[0].usb_port == USBPort(
            name=expected_name,
            port_number=9,
            port_group=PortGroup.FRONT,
            device_path="1.0/tty/ttyACM5/dev",
            hub=True,
            hub_port=4,
        )



from pathlib import Path

import pytest


def test_find_smoothie_file(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    import opentrons

    dummy_file = tmp_path / "smoothie-edge-2cac98asda.hex"
    dummy_file.write_text("hello")
    monkeypatch.setattr(opentrons, "ROBOT_FIRMWARE_DIR", tmp_path)

    monkeypatch.setattr(opentrons, "IS_ROBOT", True)
    assert opentrons._find_smoothie_file() == (dummy_file, "edge-2cac98asda")



"""Tests for the CommandStore+CommandState+CommandView trifecta.

The trifecta is tested here as a single unit, treating CommandState as a private
implementation detail.
"""

from datetime import datetime
from typing import Any
from unittest.mock import sentinel

import pytest

from opentrons_shared_data.errors import ErrorCodes, PythonException

from opentrons.hardware_control.types import DoorState
from opentrons.ordered_set import OrderedSet
from opentrons.protocol_engine import actions, commands, errors
from opentrons.protocol_engine.actions.actions import (
    PlayAction,
    SetErrorRecoveryPolicyAction,
)
from opentrons.protocol_engine.commands.command import CommandIntent, DefinedErrorData
from opentrons.protocol_engine.error_recovery_policy import ErrorRecoveryType
from opentrons.protocol_engine.errors.error_occurrence import ErrorOccurrence
from opentrons.protocol_engine.errors.exceptions import (
    EStopActivatedError,
    RobotDoorOpenError,
)
from opentrons.protocol_engine.notes.notes import CommandNote
from opentrons.protocol_engine.state.commands import (
    CommandErrorSlice,
    CommandStore,
    CommandView,
)
from opentrons.protocol_engine.state.config import Config
from opentrons.protocol_engine.state.update_types import StateUpdate
from opentrons.protocol_engine.types import DeckType, EngineStatus


def _make_config() -> Config:
    return Config(
        # Choice of robot and deck type is arbitrary.
        robot_type="OT-2 Standard",
        deck_type=DeckType.OT2_STANDARD,
    )


def _placeholder_error_recovery_policy(*args: object, **kwargs: object) -> Any:
    """A placeholder `ErrorRecoveryPolicy` for tests that don't care about it."""
    raise NotImplementedError()


def test_queue_command_action() -> None:
    """It should translate a command request into a queued command and add it."""
    subject = CommandStore(
        is_door_open=False,
        config=_make_config(),
        error_recovery_policy=_placeholder_error_recovery_policy,
    )
    subject_view = CommandView(subject.state)

    id = "command-id"
    key = "command-key"
    params = commands.CommentParams(message="yay")
    created_at = datetime(year=2021, month=1, day=1)
    request = commands.CommentCreate(params=params, key=key)
    action = actions.QueueCommandAction(
        request=request,
        request_hash=None,
        created_at=created_at,
        command_id=id,
    )
    expected_command = commands.Comment(
        id=id,
        key=key,
        createdAt=created_at,
        status=commands.CommandStatus.QUEUED,
        params=params,
    )

    subject.handle_action(action)
    assert subject_view.get("command-id") == expected_command
    assert subject_view.get_all() == [expected_command]


def test_latest_protocol_command_hash() -> None:
    """It should return the latest protocol command's hash."""
    subject = CommandStore(
        is_door_open=False,
        config=_make_config(),
        error_recovery_policy=_placeholder_error_recovery_policy,
    )
    subject_view = CommandView(subject.state)

    # The initial hash should be None.
    assert subject_view.get_latest_protocol_command_hash() is None

    # It should pick up the hash from an enqueued protocol command.
    subject.handle_action(
        actions.QueueCommandAction(
            request=commands.CommentCreate(
                params=commands.CommentParams(message="hello world"),
            ),
            request_hash="hash-1",
            command_id="command-id-1",
            created_at=datetime.now(),
        )
    )
    assert subject_view.get_latest_protocol_command_hash() == "hash-1"

    # It should pick up newer hashes as they come in.
    subject.handle_action(
        actions.QueueCommandAction(
            request=commands.CommentCreate(
                params=commands.CommentParams(message="hello world"),
            ),
            request_hash="hash-2",
            command_id="command-id-2",
            created_at=datetime.now(),
        )
    )
    assert subject_view.get_latest_protocol_command_hash() == "hash-2"


@pytest.mark.parametrize("error_recovery_type", ErrorRecoveryType)
def test_command_failure(error_recovery_type: ErrorRecoveryType) -> None:
    """It should store an error and mark the command if it fails."""
    subject = CommandStore(
        is_door_open=False,
        config=_make_config(),
        error_recovery_policy=_placeholder_error_recovery_policy,
    )
    subject_view = CommandView(subject.state)

    command_id = "command-id"
    command_key = "command-key"
    created_at = datetime(year=2021, month=1, day=1)
    started_at = datetime(year=2022, month=2, day=2)
    failed_at = datetime(year=2023, month=3, day=3)
    error_id = "error-id"
    notes = [
        CommandNote(
            noteKind="noteKind",
            shortMessage="shortMessage",
            longMessage="longMessage",
            source="source",
        )
    ]

    params = commands.CommentParams(message="No comment.")

    subject.handle_action(
        actions.QueueCommandAction(
            command_id=command_id,
            created_at=created_at,
            request=commands.CommentCreate(params=params, key=command_key),
            request_hash=None,
        )
    )
    subject.handle_action(
        actions.RunCommandAction(command_id=command_id, started_at=started_at)
    )
    subject.handle_action(
        actions.FailCommandAction(
            command_id=command_id,
            running_command=subject_view.get(command_id),
            error_id=error_id,
            failed_at=failed_at,
            error=errors.ProtocolEngineError(message="oh no"),
            notes=notes,
            type=error_recovery_type,
        )
    )

    expected_error_occurrence = errors.ErrorOccurrence(
        id=error_id,
        errorType="ProtocolEngineError",
        createdAt=failed_at,
        detail="oh no",
        errorCode=ErrorCodes.GENERAL_ERROR.value.code,
    )
    expected_failed_command = commands.Comment(
        id=command_id,
        key=command_key,
        commandType="comment",
        createdAt=created_at,
        startedAt=started_at,
        completedAt=failed_at,
        status=commands.CommandStatus.FAILED,
        params=params,
        result=None,
        error=expected_error_occurrence,
        notes=notes,
    )

    assert subject_view.get("command-id") == expected_failed_command
    assert subject_view.get_all_errors() == [expected_error_occurrence]


def test_command_failure_clears_queues() -> None:
    """It should clear the command queue on command failure."""
    subject = CommandStore(
        config=_make_config(),
        is_door_open=False,
        error_recovery_policy=_placeholder_error_recovery_policy,
    )
    subject_view = CommandView(subject.state)

    queue_1 = actions.QueueCommandAction(
        request=commands.WaitForResumeCreate(
            params=commands.WaitForResumeParams(), key="command-key-1"
        ),
        request_hash=None,
        created_at=datetime(year=2021, month=1, day=1),
        command_id="command-id-1",
    )
    subject.handle_action(queue_1)
    queue_2 = actions.QueueCommandAction(
        request=commands.WaitForResumeCreate(
            params=commands.WaitForResumeParams(), key="command-key-2"
        ),
        request_hash=None,
        created_at=datetime(year=2021, month=1, day=1),
        command_id="command-id-2",
    )
    subject.handle_action(queue_2)

    run_1 = actions.RunCommandAction(
        command_id="command-id-1",
        started_at=datetime(year=2022, month=2, day=2),
    )
    subject.handle_action(run_1)
    expected_error = errors.ProtocolEngineError(message="oh no")
    expected_error_occurance = errors.ErrorOccurrence(
        id="error-id",
        errorType="ProtocolEngineError",
        createdAt=datetime(year=2023, month=3, day=3),
        detail="oh no",
        errorCode=ErrorCodes.GENERAL_ERROR.value.code,
    )
    fail_1 = actions.FailCommandAction(
        command_id="command-id-1",
        running_command=subject_view.get("command-id-1"),
        error_id="error-id",
        failed_at=datetime(year=2023, month=3, day=3),
        error=expected_error,
        notes=[],
        type=ErrorRecoveryType.FAIL_RUN,
    )
    subject.handle_action(fail_1)

    assert [(c.id, c.status) for c in subject_view.get_all()] == [
        ("command-id-1", commands.CommandStatus.FAILED),
        ("command-id-2", commands.CommandStatus.FAILED),
    ]
    assert subject_view.get_running_command_id() is None
    assert subject_view.get_queue_ids() == OrderedSet()
    assert subject_view.get_next_to_execute() is None
    assert subject_view.get_all_errors() == [expected_error_occurance]


def test_setup_command_failure_only_clears_setup_command_queue() -> None:
    """It should clear only the setup command queue for a failed setup command.

    This test queues up a non-setup command followed by two setup commands,
    then runs and fails the first setup command.
    """
    subject = CommandStore(
        is_door_open=False,
        config=_make_config(),
        error_recovery_policy=_placeholder_error_recovery_policy,
    )
    subject_view = CommandView(subject.state)

    queue_1 = actions.QueueCommandAction(
        request=commands.WaitForResumeCreate(
            params=commands.WaitForResumeParams(), key="command-key-1"
        ),
        request_hash=None,
        created_at=datetime(year=2021, month=1, day=1),
        command_id="command-id-1",
    )
    subject.handle_action(queue_1)
    queue_2_setup = actions.QueueCommandAction(
        request=commands.WaitForResumeCreate(
            params=commands.WaitForResumeParams(),
            intent=commands.CommandIntent.SETUP,
            key="command-key-2",
        ),
        request_hash=None,
        created_at=datetime(year=2021, month=1, day=1),
        command_id="command-id-2",
    )
    subject.handle_action(queue_2_setup)
    queue_3_setup = actions.QueueCommandAction(
        request=commands.WaitForResumeCreate(
            params=commands.WaitForResumeParams(),
            intent=commands.CommandIntent.SETUP,
            key="command-key-3",
        ),
        request_hash=None,
        created_at=datetime(year=2021, month=1, day=1),
        command_id="command-id-3",
    )
    subject.handle_action(queue_3_setup)

    run_2_setup = actions.RunCommandAction(
        command_id="command-id-2",
        started_at=datetime(year=2022, month=2, day=2),
    )
    subject.handle_action(run_2_setup)
    fail_2_setup = actions.FailCommandAction(
        command_id="command-id-2",
        running_command=subject_view.get("command-id-2"),
        error_id="error-id",
        failed_at=datetime(year=2023, month=3, day=3),
        error=errors.ProtocolEngineError(message="oh no"),
        notes=[],
        type=ErrorRecoveryType.FAIL_RUN,
    )
    subject.handle_action(fail_2_setup)

    assert [(c.id, c.status) for c in subject_view.get_all()] == [
        ("command-id-1", commands.CommandStatus.QUEUED),
        ("command-id-2", commands.CommandStatus.FAILED),
        ("command-id-3", commands.CommandStatus.FAILED),
    ]
    assert subject_view.get_running_command_id() is None

    subject.handle_action(actions.PlayAction(requested_at=datetime.now()))
    assert subject_view.get_next_to_execute() == "command-id-1"


def test_nonfatal_command_failure() -> None:
    """Test the command queue if a command fails recoverably.

    Commands that were after the failed command in the queue should be left in
    the queue.

    The queue status should be "awaiting-recovery."
    """
    subject = CommandStore(
        is_door_open=False,
        config=_make_config(),
        error_recovery_policy=_placeholder_error_recovery_policy,
    )
    subject_view = CommandView(subject.state)

    queue_1 = actions.QueueCommandAction(
        request=commands.WaitForResumeCreate(
            params=commands.WaitForResumeParams(), key="command-key-1"
        ),
        request_hash=None,
        created_at=datetime(year=2021, month=1, day=1),
        command_id="command-id-1",
    )
    subject.handle_action(queue_1)
    queue_2 = actions.QueueCommandAction(
        request=commands.WaitForResumeCreate(
            params=commands.WaitForResumeParams(), key="command-key-2"
        ),
        request_hash=None,
        created_at=datetime(year=2021, month=1, day=1),
        command_id="command-id-2",
    )
    subject.handle_action(queue_2)

    run_1 = actions.RunCommandAction(
        command_id="command-id-1",
        started_at=datetime(year=2022, month=2, day=2),
    )
    subject.handle_action(run_1)
    fail_1 = actions.FailCommandAction(
        command_id="command-id-1",
        running_command=subject_view.get("command-id-1"),
        error_id="error-id",
        failed_at=datetime(year=2023, month=3, day=3),
        error=errors.ProtocolEngineError(message="oh no"),
        notes=[],
        type=ErrorRecoveryType.WAIT_FOR_RECOVERY,
    )
    subject.handle_action(fail_1)

    assert subject_view.get_status() == EngineStatus.AWAITING_RECOVERY
    assert [(c.id, c.status) for c in subject_view.get_all()] == [
        ("command-id-1", commands.CommandStatus.FAILED),
        ("command-id-2", commands.CommandStatus.QUEUED),
    ]
    assert subject_view.get_running_command_id() is None


def test_nonfatal_command_failure_with_door_open() -> None:
    """Test the command queue if a command fails recoverably but the door is open.

    Commands that were after the failed command in the queue should be left in
    the queue.

    The queue status should be "awaiting-recovery-paused."
    """
    subject = CommandStore(
        is_door_open=False,
        config=Config(
            block_on_door_open=True,
            # Choice of robot and deck type are arbitrary.
            robot_type="OT-3 Standard",
            deck_type=DeckType.OT3_STANDARD,
        ),
        error_recovery_policy=_placeholder_error_recovery_policy,
    )
    subject_view = CommandView(subject.state)

    queue_1 = actions.QueueCommandAction(
        request=commands.WaitForResumeCreate(
            params=commands.WaitForResumeParams(), key="command-key-1"
        ),
        request_hash=None,
        created_at=datetime(year=2021, month=1, day=1),
        command_id="command-id-1",
    )
    subject.handle_action(queue_1)
    queue_2 = actions.QueueCommandAction(
        request=commands.WaitForResumeCreate(
            params=commands.WaitForResumeParams(), key="command-key-2"
        ),
        request_hash=None,
        created_at=datetime(year=2021, month=1, day=1),
        command_id="command-id-2",
    )
    subject.handle_action(queue_2)

    run_1 = actions.RunCommandAction(
        command_id="command-id-1",
        started_at=datetime(year=2022, month=2, day=2),
    )
    subject.handle_action(run_1)
    door_open = actions.DoorChangeAction(
        door_state=DoorState.OPEN,
        module_serial=None,
    )
    subject.handle_action(door_open)
    assert subject_view.get_is_door_blocking() is True

    fail_1 = actions.FailCommandAction(
        command_id="command-id-1",
        running_command=subject_view.get("command-id-1"),
        error_id="error-id",
        failed_at=datetime(year=2023, month=3, day=3),
        error=errors.ProtocolEngineError(message="oh no"),
        notes=[],
        type=ErrorRecoveryType.WAIT_FOR_RECOVERY,
    )
    subject.handle_action(fail_1)

    assert (
        subject_view.get_status() == EngineStatus.AWAITING_RECOVERY_BLOCKED_BY_OPEN_DOOR
    )
    assert [(c.id, c.status) for c in subject_view.get_all()] == [
        ("command-id-1", commands.CommandStatus.FAILED),
        ("command-id-2", commands.CommandStatus.QUEUED),
    ]
    assert subject_view.get_running_command_id() is None


def test_fixit_command_failure_handling_by_recovery_type() -> None:
    """Test that fixit command failures are handled differently based on error recovery type.

    When a fixit command fails with permissible error recovery types, other queued fixit
    commands should remain in the queue as is.

    When a fixit command fails with other error recovery types, all queued fixit commands
    should be marked as failed.
    """

    def test_with_recovery_type(
        recovery_type: ErrorRecoveryType, should_fail_queued_cmds: bool
    ) -> None:
        subject = CommandStore(
            is_door_open=False,
            config=_make_config(),
            error_recovery_policy=_placeholder_error_recovery_policy,
        )
        subject_view = CommandView(subject.state)

        protocol_cmd = actions.QueueCommandAction(
            request=commands.CommentCreate(
                params=commands.CommentParams(message=""),
                key="protocol-cmd",
            ),
            request_hash=None,
            created_at=datetime(year=2021, month=1, day=1),
            command_id="protocol-cmd-id",
        )
        subject.handle_action(protocol_cmd)
        subject.handle_action(
            actions.RunCommandAction(
                command_id="protocol-cmd-id",
                started_at=datetime(year=2021, month=1, day=2),
            )
        )
        subject.handle_action(
            actions.FailCommandAction(
                command_id="protocol-cmd-id",
                running_command=subject_view.get("protocol-cmd-id"),
                error_id="error-1",
                failed_at=datetime(year=2021, month=1, day=3),
                error=errors.ProtocolEngineError(message="recovery needed"),
                notes=[],
                type=ErrorRecoveryType.WAIT_FOR_RECOVERY,
            )
        )

        for i in range(1, 4):
            subject.handle_action(
                actions.QueueCommandAction(
                    request=commands.CommentCreate(
                        params=commands.CommentParams(message=f"fixit {i}"),
                        key=f"fixit-{i}",
                        intent=commands.CommandIntent.FIXIT,
                    ),
                    request_hash=None,
                    created_at=datetime(year=2021, month=2, day=i),
                    command_id=f"fixit-id-{i}",
                )
            )

        subject.handle_action(
            actions.RunCommandAction(
                command_id="fixit-id-1",
                started_at=datetime(year=2021, month=2, day=10),
            )
        )
        subject.handle_action(
            actions.FailCommandAction(
                command_id="fixit-id-1",
                running_command=subject_view.get("fixit-id-1"),
                error_id="error-2",
                failed_at=datetime(year=2021, month=2, day=11),
                error=errors.ProtocolEngineError(
                    message=f"fixit failed with {recovery_type}"
                ),
                notes=[],
                type=recovery_type,
            )
        )

        if should_fail_queued_cmds:
            assert all(
                c.status == commands.CommandStatus.FAILED
                for c in subject_view.get_all()
            )
        else:
            assert [(c.id, c.status) for c in subject_view.get_all()] == [
                ("protocol-cmd-id", commands.CommandStatus.FAILED),
                ("fixit-id-1", commands.CommandStatus.FAILED),
                ("fixit-id-2", commands.CommandStatus.QUEUED),
                ("fixit-id-3", commands.CommandStatus.QUEUED),
            ]
            assert subject_view.get_next_to_execute() == "fixit-id-2"

    test_with_recovery_type(
        ErrorRecoveryType.CONTINUE_WITH_ERROR, should_fail_queued_cmds=False
    )
    test_with_recovery_type(
        ErrorRecoveryType.ASSUME_FALSE_POSITIVE_AND_CONTINUE,
        should_fail_queued_cmds=False,
    )
    test_with_recovery_type(ErrorRecoveryType.FAIL_RUN, should_fail_queued_cmds=True)


def test_door_during_setup_phase() -> None:
    """Test behavior when the door is opened during the setup phase."""
    subject = CommandStore(
        is_door_open=False,
        error_recovery_policy=_placeholder_error_recovery_policy,
        config=Config(
            block_on_door_open=True,
            # Choice of robot and deck type are arbitrary.
            robot_type="OT-2 Standard",
            deck_type=DeckType.OT2_STANDARD,
        ),
    )
    subject_view = CommandView(subject.state)

    queue_setup_command = actions.QueueCommandAction(
        request=commands.CommentCreate(
            params=commands.CommentParams(message=""),
            key="setup-command-key",
            intent=commands.CommandIntent.SETUP,
        ),
        request_hash=None,
        created_at=datetime(year=2021, month=1, day=1),
        command_id="setup-command-id",
    )
    subject.handle_action(queue_setup_command)

    subject.handle_action(actions.DoorChangeAction(DoorState.OPEN))

    assert subject_view.get_status() == EngineStatus.IDLE
    # The door being open should not block the setup command from executing.
    assert subject_view.get_next_to_execute() == "setup-command-id"


def test_door_during_protocol_phase() -> None:
    """Test behavior when the door is opened during the main protocol phase."""
    subject = CommandStore(
        is_door_open=False,
        error_recovery_policy=_placeholder_error_recovery_policy,
        config=Config(
            block_on_door_open=True,
            # Choice of robot and deck type are arbitrary.
            robot_type="OT-2 Standard",
            deck_type=DeckType.OT2_STANDARD,
        ),
    )
    subject_view = CommandView(subject.state)

    queue_protocol_command = actions.QueueCommandAction(
        request=commands.CommentCreate(
            params=commands.CommentParams(message=""),
            key="command-key",
        ),
        request_hash=None,
        created_at=datetime(year=2021, month=1, day=1),
        command_id="command-id",
    )
    subject.handle_action(queue_protocol_command)

    subject.handle_action(actions.PlayAction(requested_at=datetime.now()))

    play = PlayAction(requested_at=datetime.now())

    # Test state after we open the door:
    subject.handle_action(actions.DoorChangeAction(DoorState.OPEN))
    assert subject_view.get_status() == EngineStatus.BLOCKED_BY_OPEN_DOOR
    assert subject_view.get_next_to_execute() is None
    with pytest.raises(RobotDoorOpenError):
        subject_view.validate_action_allowed(play)

    # Test state after we close the door (with an extra open-close for good measure)
    subject.handle_action(actions.DoorChangeAction(DoorState.CLOSED))
    subject.handle_action(actions.DoorChangeAction(DoorState.OPEN))
    subject.handle_action(actions.DoorChangeAction(DoorState.CLOSED))
    assert subject_view.get_status() == EngineStatus.PAUSED
    assert subject_view.get_next_to_execute() is None
    subject_view.validate_action_allowed(play)  # Should not raise.

    # Test state when we resume:
    subject.handle_action(play)
    assert subject_view.get_status() == EngineStatus.RUNNING
    assert subject_view.get_next_to_execute() == "command-id"


def test_door_during_error_recovery() -> None:
    """Test behavior when the door is opened during error recovery."""
    subject = CommandStore(
        is_door_open=False,
        error_recovery_policy=_placeholder_error_recovery_policy,
        config=Config(
            block_on_door_open=True,
            # Choice of robot and deck type are arbitrary.
            robot_type="OT-2 Standard",
            deck_type=DeckType.OT2_STANDARD,
        ),
    )
    subject_view = CommandView(subject.state)

    # Fail a command to put the subject in recovery mode.
    queue_1 = actions.QueueCommandAction(
        request=commands.CommentCreate(
            params=commands.CommentParams(message=""), key="command-key-1"
        ),
        request_hash=None,
        created_at=datetime(year=2021, month=1, day=1),
        command_id="command-id-1",
    )
    subject.handle_action(queue_1)
    run_1 = actions.RunCommandAction(
        command_id="command-id-1",
        started_at=datetime(year=2022, month=2, day=2),
    )
    subject.handle_action(run_1)
    expected_error = errors.ProtocolEngineError(message="oh no")
    expected_error_occurance = errors.ErrorOccurrence(
        id="error-id",
        errorType="ProtocolEngineError",
        createdAt=datetime(year=2023, month=3, day=3),
        detail="oh no",
        errorCode=ErrorCodes.GENERAL_ERROR.value.code,
    )
    fail_1 = actions.FailCommandAction(
        command_id="command-id-1",
        running_command=subject_view.get("command-id-1"),
        error_id="error-id",
        failed_at=datetime(year=2023, month=3, day=3),
        error=expected_error,
        notes=[],
        type=ErrorRecoveryType.WAIT_FOR_RECOVERY,
    )
    subject.handle_action(fail_1)

    queue_2 = actions.QueueCommandAction(
        request=commands.CommentCreate(
            params=commands.CommentParams(message=""),
            key="command-key-2",
            intent=CommandIntent.FIXIT,
        ),
        request_hash=None,
        created_at=datetime(year=2021, month=1, day=1),
        command_id="command-id-2",
    )
    subject.handle_action(queue_2)
    assert subject_view.get_status() == EngineStatus.AWAITING_RECOVERY
    assert subject_view.get_next_to_execute() == "command-id-2"

    # Test state after we open the door:
    subject.handle_action(actions.DoorChangeAction(DoorState.OPEN))
    assert (
        subject_view.get_status() == EngineStatus.AWAITING_RECOVERY_BLOCKED_BY_OPEN_DOOR
    )
    assert subject_view.get_next_to_execute() is None
    play = actions.PlayAction(requested_at=datetime.now())
    with pytest.raises(RobotDoorOpenError):
        subject_view.validate_action_allowed(play)

    # Test state after we close the door (with an extra open-close for good measure)
    subject.handle_action(actions.DoorChangeAction(DoorState.CLOSED))
    subject.handle_action(actions.DoorChangeAction(DoorState.OPEN))
    subject.handle_action(actions.DoorChangeAction(DoorState.CLOSED))
    assert subject_view.get_status() == EngineStatus.AWAITING_RECOVERY_PAUSED
    assert subject_view.get_next_to_execute() is None
    subject_view.validate_action_allowed(play)  # Should not raise.

    # Test state when we resume recovery mode:
    subject.handle_action(play)
    assert subject_view.get_status() == EngineStatus.AWAITING_RECOVERY
    assert subject_view.get_next_to_execute() == "command-id-2"
    assert subject_view.get_all_errors() == [expected_error_occurance]


@pytest.mark.parametrize("close_door_before_queueing", [False, True])
def test_door_ungrip_labware(close_door_before_queueing: bool) -> None:
    """Ungrip commands should be able to run even when the door is open."""
    subject = CommandStore(
        is_door_open=False,
        error_recovery_policy=_placeholder_error_recovery_policy,
        config=Config(
            block_on_door_open=True,
            # Choice of robot and deck type are arbitrary.
            robot_type="OT-2 Standard",
            deck_type=DeckType.OT2_STANDARD,
        ),
    )
    subject_view = CommandView(subject.state)

    # Fail a command to put the subject in recovery mode.
    queue_failing = actions.QueueCommandAction(
        request=commands.CommentCreate(
            params=commands.CommentParams(message=""), key="command-key-1"
        ),
        request_hash=None,
        created_at=datetime(year=2021, month=1, day=1),
        command_id="failing-command-id",
    )
    subject.handle_action(queue_failing)
    run_failing = actions.RunCommandAction(
        command_id="failing-command-id",
        started_at=datetime(year=2022, month=2, day=2),
    )
    subject.handle_action(run_failing)
    expected_error = errors.ProtocolEngineError(message="oh no")
    fail_failing = actions.FailCommandAction(
        command_id="failing-command-id",
        running_command=subject_view.get("failing-command-id"),
        error_id="error-id",
        failed_at=datetime(year=2023, month=3, day=3),
        error=expected_error,
        notes=[],
        type=ErrorRecoveryType.WAIT_FOR_RECOVERY,
    )
    subject.handle_action(fail_failing)

    # Open the door:
    subject.handle_action(actions.DoorChangeAction(DoorState.OPEN))
    assert (
        subject_view.get_status() == EngineStatus.AWAITING_RECOVERY_BLOCKED_BY_OPEN_DOOR
    )
    assert subject_view.get_next_to_execute() is None

    if close_door_before_queueing:
        subject.handle_action(actions.DoorChangeAction(DoorState.CLOSED))

    assert subject_view.get_status() in (
        EngineStatus.AWAITING_RECOVERY_PAUSED,  # If we closed the door.
        EngineStatus.AWAITING_RECOVERY_BLOCKED_BY_OPEN_DOOR,  # If we didn't.
    )

    # Make sure the special ungrip command can be queued and that it will be returned
    # as next to execute:
    queue_fixit = actions.QueueCommandAction(
        request=commands.unsafe.UnsafeUngripLabwareCreate(
            params=commands.unsafe.UnsafeUngripLabwareParams(),
            intent=CommandIntent.FIXIT,
        ),
        request_hash=None,
        created_at=datetime(year=2021, month=1, day=1),
        command_id="fixit-command-id",
    )
    subject_view.validate_action_allowed(queue_fixit)
    subject.handle_action(queue_fixit)
    assert subject_view.get_next_to_execute() == "fixit-command-id"


@pytest.mark.parametrize(
    ("door_initially_open", "expected_engine_status_after_play"),
    [
        (False, EngineStatus.RUNNING),
        (True, EngineStatus.BLOCKED_BY_OPEN_DOOR),
    ],
)
def test_door_initially_open(
    door_initially_open: bool, expected_engine_status_after_play: EngineStatus
) -> None:
    """Test open-door blocking behavior given different initial door states."""
    subject = CommandStore(
        is_door_open=door_initially_open,
        error_recovery_policy=_placeholder_error_recovery_policy,
        config=Config(
            block_on_door_open=True,
            # Choice of robot and deck type are arbitrary.
            robot_type="OT-2 Standard",
            deck_type=DeckType.OT2_STANDARD,
        ),
    )
    subject_view = CommandView(subject.state)

    assert subject_view.get_status() == EngineStatus.IDLE
    play = actions.PlayAction(requested_at=datetime.now())
    subject_view.validate_action_allowed(play)  # Should not raise.
    subject.handle_action(play)
    assert subject_view.get_status() == expected_engine_status_after_play


def test_error_recovery_type_tracking() -> None:
    """It should keep track of each failed command's error recovery type."""
    subject = CommandStore(
        config=_make_config(),
        error_recovery_policy=_placeholder_error_recovery_policy,
        is_door_open=False,
    )

    subject.handle_action(
        actions.QueueCommandAction(
            command_id="c1",
            created_at=datetime.now(),
            request=commands.CommentCreate(
                params=commands.CommentParams(message="yeehaw"),
            ),
            request_hash=None,
        )
    )
    subject.handle_action(
        actions.QueueCommandAction(
            command_id="c2",
            created_at=datetime.now(),
            request=commands.CommentCreate(
                params=commands.CommentParams(message="yeehaw"),
            ),
            request_hash=None,
        )
    )
    subject.handle_action(
        actions.RunCommandAction(command_id="c1", started_at=datetime.now())
    )
    running_command_1 = CommandView(subject.state).get("c1")
    subject.handle_action(
        actions.FailCommandAction(
            command_id="c1",
            running_command=running_command_1,
            error_id="c1-error",
            failed_at=datetime(year=2023, month=3, day=3),
            error=PythonException(RuntimeError("new sheriff in town")),
            notes=[],
            type=ErrorRecoveryType.WAIT_FOR_RECOVERY,
        )
    )
    subject.handle_action(
        actions.RunCommandAction(command_id="c2", started_at=datetime.now())
    )
    running_command_2 = CommandView(subject.state).get("c2")
    subject.handle_action(
        actions.FailCommandAction(
            command_id="c2",
            running_command=running_command_2,
            error_id="c2-error",
            failed_at=datetime(year=2023, month=3, day=3),
            error=PythonException(RuntimeError("new sheriff in town")),
            notes=[],
            type=ErrorRecoveryType.FAIL_RUN,
        )
    )

    view = CommandView(subject.state)
    assert view.get_error_recovery_type("c1") == ErrorRecoveryType.WAIT_FOR_RECOVERY
    assert view.get_error_recovery_type("c2") == ErrorRecoveryType.FAIL_RUN

    exception = PythonException(RuntimeError("new sheriff in town"))
    error_occurrence_1 = ErrorOccurrence.from_failed(
        id="c1-error", createdAt=datetime(year=2023, month=3, day=3), error=exception
    )
    error_occurrence_2 = ErrorOccurrence.from_failed(
        id="c2-error", createdAt=datetime(year=2023, month=3, day=3), error=exception
    )

    assert view.get_all_errors() == [
        error_occurrence_1,
        error_occurrence_2,
    ]


def test_recovery_target_tracking() -> None:
    """It should keep track of the command currently undergoing error recovery."""
    subject = CommandStore(
        config=_make_config(),
        error_recovery_policy=_placeholder_error_recovery_policy,
        is_door_open=False,
    )
    subject_view = CommandView(subject.state)

    queue_1 = actions.QueueCommandAction(
        "c1",
        created_at=datetime.now(),
        request=commands.CommentCreate(params=commands.CommentParams(message="")),
        request_hash=None,
    )
    subject.handle_action(queue_1)
    run_1 = actions.RunCommandAction(command_id="c1", started_at=datetime.now())
    subject.handle_action(run_1)
    fail_1 = actions.FailCommandAction(
        command_id="c1",
        error_id="c1-error",
        failed_at=datetime.now(),
        error=PythonException(RuntimeError()),
        notes=[],
        type=ErrorRecoveryType.WAIT_FOR_RECOVERY,
        running_command=subject_view.get("c1"),
    )
    subject.handle_action(fail_1)

    # c1 failed recoverably and we're currently recovering from it.
    recovery_target = subject_view.get_recovery_target()
    assert recovery_target is not None
    assert recovery_target.command_id == "c1"
    assert subject_view.get_recovery_in_progress_for_command("c1")

    resume_from_1_recovery = actions.ResumeFromRecoveryAction(StateUpdate())
    subject.handle_action(resume_from_1_recovery)

    # c1 failed recoverably, but we've already completed its recovery.
    assert subject_view.get_recovery_target() is None
    assert not subject_view.get_recovery_in_progress_for_command("c1")

    queue_2 = actions.QueueCommandAction(
        "c2",
        created_at=datetime.now(),
        request=commands.CommentCreate(params=commands.CommentParams(message="")),
        request_hash=None,
    )
    subject.handle_action(queue_2)
    run_2 = actions.RunCommandAction(command_id="c2", started_at=datetime.now())
    subject.handle_action(run_2)
    fail_2 = actions.FailCommandAction(
        command_id="c2",
        error_id="c2-error",
        failed_at=datetime.now(),
        error=PythonException(RuntimeError()),
        notes=[],
        type=ErrorRecoveryType.WAIT_FOR_RECOVERY,
        running_command=subject_view.get("c2"),
    )
    subject.handle_action(fail_2)

    # c2 failed recoverably and we're currently recovering from it.
    recovery_target = subject_view.get_recovery_target()
    assert recovery_target is not None
    assert recovery_target.command_id == "c2"
    assert subject_view.get_recovery_in_progress_for_command("c2")
    # ...and that means we're *not* currently recovering from c1,
    # even though it failed recoverably before.
    assert not subject_view.get_recovery_in_progress_for_command("c1")

    resume_from_2_recovery = actions.ResumeFromRecoveryAction(StateUpdate())
    subject.handle_action(resume_from_2_recovery)
    queue_3 = actions.QueueCommandAction(
        "c3",
        created_at=datetime.now(),
        request=commands.CommentCreate(params=commands.CommentParams(message="")),
        request_hash=None,
    )
    subject.handle_action(queue_3)
    run_3 = actions.RunCommandAction(command_id="c3", started_at=datetime.now())
    subject.handle_action(run_3)
    fail_3 = actions.FailCommandAction(
        command_id="c3",
        error_id="c3-error",
        failed_at=datetime.now(),
        error=PythonException(RuntimeError()),
        notes=[],
        type=ErrorRecoveryType.FAIL_RUN,
        running_command=subject_view.get("c3"),
    )
    subject.handle_action(fail_3)

    # c3 failed, but not recoverably.
    assert subject_view.get_recovery_target() is None
    assert not subject_view.get_recovery_in_progress_for_command("c3")

    assert subject_view.get_has_entered_recovery_mode() is True


@pytest.mark.parametrize(
    "ending_action",
    [
        actions.StopAction(from_asynchronous_error=False),
        actions.StopAction(from_asynchronous_error=True),
        actions.FinishAction(set_run_status=False),
        actions.FinishAction(
            set_run_status=True,
            error_details=actions.FinishErrorDetails(
                error=Exception("blimey"),
                error_id="error-id",
                created_at=datetime.now(),
            ),
        ),
    ],
)
def test_recovery_target_clears_when_run_ends(ending_action: actions.Action) -> None:
    """There should never be an error recovery target when the run is done."""
    subject = CommandStore(
        config=_make_config(),
        error_recovery_policy=_placeholder_error_recovery_policy,
        is_door_open=False,
    )
    subject_view = CommandView(subject.state)

    # Setup: Put the run in error recovery mode.
    queue = actions.QueueCommandAction(
        "c1",
        created_at=datetime.now(),
        request=commands.CommentCreate(params=commands.CommentParams(message="")),
        request_hash=None,
    )
    subject.handle_action(queue)
    run = actions.RunCommandAction(command_id="c1", started_at=datetime.now())
    subject.handle_action(run)
    fail = actions.FailCommandAction(
        command_id="c1",
        error_id="c1-error",
        failed_at=datetime.now(),
        error=PythonException(RuntimeError()),
        notes=[],
        type=ErrorRecoveryType.WAIT_FOR_RECOVERY,
        running_command=subject_view.get("c1"),
    )
    subject.handle_action(fail)

    # Test: Assert that the ending action clears the recovery target.
    assert subject_view.get_recovery_target() is not None
    subject.handle_action(ending_action)
    assert subject_view.get_recovery_target() is None


def test_final_state_after_estop() -> None:
    """Test the final state of the run after it's E-stopped."""
    subject = CommandStore(
        config=_make_config(),
        error_recovery_policy=_placeholder_error_recovery_policy,
        is_door_open=False,
    )
    subject_view = CommandView(subject.state)

    error_details = actions.FinishErrorDetails(
        error=EStopActivatedError(), error_id="error-id", created_at=datetime.now()
    )
    expected_error_occurrence = ErrorOccurrence(
        id=error_details.error_id,
        createdAt=error_details.created_at,
        errorCode=ErrorCodes.E_STOP_ACTIVATED.value.code,
        errorType="EStopActivatedError",
        detail="E-stop activated.",
    )

    subject.handle_action(actions.StopAction(from_asynchronous_error=True))
    subject.handle_action(actions.FinishAction(error_details=error_details))
    subject.handle_action(
        actions.HardwareStoppedAction(
            completed_at=sentinel.hardware_stopped_action_completed_at,
            finish_error_details=None,
        )
    )

    assert subject_view.get_status() == EngineStatus.FAILED
    assert subject_view.get_error() == expected_error_occurrence
    assert subject_view.get_all_errors() == []


def test_final_state_after_stop() -> None:
    """Test the final state of the run after it's stopped."""
    subject = CommandStore(
        config=_make_config(),
        error_recovery_policy=_placeholder_error_recovery_policy,
        is_door_open=False,
    )
    subject_view = CommandView(subject.state)

    subject.handle_action(actions.StopAction())
    subject.handle_action(
        actions.FinishAction(
            error_details=actions.FinishErrorDetails(
                error=RuntimeError(
                    "uh oh I was a command and then I got cancelled because someone"
                    " stopped the run, and now I'm raising this exception because"
                    " of that. Woe is me"
                ),
                error_id="error-id",
                created_at=datetime.now(),
            )
        )
    )
    subject.handle_action(
        actions.HardwareStoppedAction(
            completed_at=sentinel.hardware_stopped_action_completed_at,
            finish_error_details=None,
        )
    )

    assert subject_view.get_status() == EngineStatus.STOPPED
    assert subject_view.get_error() is None


def test_final_state_after_error_recovery_stop() -> None:
    """Test the final state of the run after it's stopped during error recovery.

    We still want to count this as "stopped," not "failed."
    """
    subject = CommandStore(
        config=_make_config(),
        error_recovery_policy=_placeholder_error_recovery_policy,
        is_door_open=False,
    )
    subject_view = CommandView(subject.state)

    # Fail a command to put the subject in recovery mode.
    queue_1 = actions.QueueCommandAction(
        request=commands.CommentCreate(
            params=commands.CommentParams(message=""), key="command-key-1"
        ),
        request_hash=None,
        created_at=datetime(year=2021, month=1, day=1),
        command_id="command-id-1",
    )
    subject.handle_action(queue_1)
    run_1 = actions.RunCommandAction(
        command_id="command-id-1",
        started_at=datetime(year=2022, month=2, day=2),
    )
    subject.handle_action(run_1)
    fail_1 = actions.FailCommandAction(
        command_id="command-id-1",
        running_command=subject_view.get("command-id-1"),
        error_id="error-id",
        failed_at=datetime(year=2023, month=3, day=3),
        error=errors.ProtocolEngineError(message="oh no"),
        notes=[],
        type=ErrorRecoveryType.WAIT_FOR_RECOVERY,
    )
    expected_error_occurrence_1 = ErrorOccurrence(
        id="error-id",
        createdAt=datetime(year=2023, month=3, day=3),
        errorCode=ErrorCodes.GENERAL_ERROR.value.code,
        errorType="ProtocolEngineError",
        detail="oh no",
    )
    subject.handle_action(fail_1)
    assert subject_view.get_status() == EngineStatus.AWAITING_RECOVERY

    subject.handle_action(actions.StopAction())
    subject.handle_action(
        actions.FinishAction(
            error_details=actions.FinishErrorDetails(
                error=RuntimeError(
                    "uh oh I was a command and then I got cancelled because someone"
                    " stopped the run, and now I'm raising this exception because"
                    " of that. Woe is me"
                ),
                error_id="error-id",
                created_at=datetime.now(),
            )
        )
    )
    subject.handle_action(
        actions.HardwareStoppedAction(
            completed_at=sentinel.hardware_stopped_action_completed_at,
            finish_error_details=None,
        )
    )

    assert subject_view.get_status() == EngineStatus.STOPPED
    assert subject_view.get_recovery_target() is None
    assert subject_view.get_error() is None
    assert subject_view.get_all_errors() == [
        expected_error_occurrence_1,
    ]


def test_set_and_get_error_recovery_policy() -> None:
    """Test storage of `ErrorRecoveryPolicy`s."""
    initial_policy = sentinel.initial_policy
    new_policy = sentinel.new_policy
    subject = CommandStore(
        config=_make_config(),
        error_recovery_policy=initial_policy,
        is_door_open=False,
    )
    subject_view = CommandView(subject.state)
    assert subject_view.get_error_recovery_policy() is initial_policy
    subject.handle_action(SetErrorRecoveryPolicyAction(sentinel.new_policy))
    assert subject_view.get_error_recovery_policy() is new_policy


def test_get_state_update_for_false_positive() -> None:
    """Test storage of false-positive state updates."""
    subject = CommandStore(
        config=_make_config(),
        error_recovery_policy=_placeholder_error_recovery_policy,
        is_door_open=False,
    )
    subject_view = CommandView(subject.state)

    empty_state_update = StateUpdate()

    assert subject_view.get_state_update_for_false_positive() == empty_state_update

    queue = actions.QueueCommandAction(
        request=commands.CommentCreate(
            params=commands.CommentParams(message=""), key="command-key-1"
        ),
        request_hash=None,
        created_at=datetime(year=2021, month=1, day=1),
        command_id="command-id-1",
    )
    subject.handle_action(queue)
    run = actions.RunCommandAction(
        command_id="command-id-1",
        started_at=datetime(year=2022, month=2, day=2),
    )
    subject.handle_action(run)
    fail = actions.FailCommandAction(
        command_id="command-id-1",
        running_command=subject_view.get("command-id-1"),
        error_id="error-id",
        failed_at=datetime(year=2023, month=3, day=3),
        error=DefinedErrorData(
            public=sentinel.public,
            state_update_if_false_positive=sentinel.state_update_if_false_positive,
        ),
        type=ErrorRecoveryType.WAIT_FOR_RECOVERY,
        notes=[],
    )
    subject.handle_action(fail)

    assert (
        subject_view.get_state_update_for_false_positive()
        == sentinel.state_update_if_false_positive
    )

    resume_from_recovery = actions.ResumeFromRecoveryAction(
        state_update=sentinel.some_other_state_update
    )
    subject.handle_action(resume_from_recovery)

    assert subject_view.get_state_update_for_false_positive() == empty_state_update


def test_get_errors_slice_empty() -> None:
    """It should return an empty error list."""
    subject = CommandStore(
        config=_make_config(),
        error_recovery_policy=_placeholder_error_recovery_policy,
        is_door_open=False,
    )
    subject_view = CommandView(subject.state)
    result = subject_view.get_errors_slice(cursor=0, length=2)

    assert result == CommandErrorSlice(commands_errors=[], cursor=0, total_length=0)


def test_get_errors_slice() -> None:
    """It should return a slice of all command errors."""
    subject = CommandStore(
        config=_make_config(),
        error_recovery_policy=_placeholder_error_recovery_policy,
        is_door_open=False,
    )

    subject_view = CommandView(subject.state)

    queue_1 = actions.QueueCommandAction(
        request=commands.WaitForResumeCreate(
            params=commands.WaitForResumeParams(), key="command-key-1"
        ),
        request_hash=None,
        created_at=datetime(year=2021, month=1, day=1),
        command_id="command-id-1",
    )
    subject.handle_action(queue_1)
    queue_2_setup = actions.QueueCommandAction(
        request=commands.WaitForResumeCreate(
            params=commands.WaitForResumeParams(),
            intent=commands.CommandIntent.SETUP,
            key="command-key-2",
        ),
        request_hash=None,
        created_at=datetime(year=2021, month=1, day=1),
        command_id="command-id-2",
    )
    subject.handle_action(queue_2_setup)
    queue_3_setup = actions.QueueCommandAction(
        request=commands.WaitForResumeCreate(
            params=commands.WaitForResumeParams(),
            intent=commands.CommandIntent.SETUP,
            key="command-key-3",
        ),
        request_hash=None,
        created_at=datetime(year=2021, month=1, day=1),
        command_id="command-id-3",
    )
    subject.handle_action(queue_3_setup)

    run_2_setup = actions.RunCommandAction(
        command_id="command-id-2",
        started_at=datetime(year=2022, month=2, day=2),
    )
    subject.handle_action(run_2_setup)
    fail_2_setup = actions.FailCommandAction(
        command_id="command-id-2",
        running_command=subject_view.get("command-id-2"),
        error_id="error-id",
        failed_at=datetime(year=2023, month=3, day=3),
        error=errors.ProtocolEngineError(message="oh no"),
        notes=[],
        type=ErrorRecoveryType.CONTINUE_WITH_ERROR,
    )
    subject.handle_action(fail_2_setup)

    result = subject_view.get_errors_slice(cursor=1, length=3)

    assert result == CommandErrorSlice(
        [
            ErrorOccurrence(
                id="error-id",
                createdAt=datetime(2023, 3, 3, 0, 0),
                isDefined=False,
                errorType="ProtocolEngineError",
                errorCode="4000",
                detail="oh no",
                errorInfo={},
                wrappedErrors=[],
            )
        ],
        cursor=0,
        total_length=1,
    )



from typing import Optional

import pytest
from mock import AsyncMock

from opentrons.drivers.asyncio.communication.serial_connection import (
    AsyncResponseSerialConnection,
)
from opentrons.drivers.command_builder import CommandBuilder
from opentrons.drivers.thermocycler import driver
from opentrons.drivers.types import PlateTemperature, Temperature, ThermocyclerLidStatus
from opentrons.drivers.utils import TC_GCODE_ROUNDING_PRECISION


@pytest.fixture
def connection() -> AsyncMock:
    return AsyncMock(spec=AsyncResponseSerialConnection)


@pytest.fixture
def subject(connection: AsyncMock) -> driver.ThermocyclerDriverV2:
    connection.send_command.return_value = ""
    return driver.ThermocyclerDriverV2(connection)


async def test_open_lid(
    subject: driver.ThermocyclerDriverV2, connection: AsyncMock
) -> None:
    """It should send an open lid command."""
    await subject.open_lid()

    expected = CommandBuilder(terminator=driver.TC_COMMAND_TERMINATOR).add_gcode(
        gcode="M126"
    )

    connection.send_command.assert_called_once_with(command=expected, retries=3)


async def test_close_lid(
    subject: driver.ThermocyclerDriverV2, connection: AsyncMock
) -> None:
    """It should send a close lid command."""
    await subject.close_lid()

    expected = CommandBuilder(terminator=driver.TC_COMMAND_TERMINATOR).add_gcode(
        gcode="M127"
    )

    connection.send_command.assert_called_once_with(command=expected, retries=3)


async def test_plate_lift(
    subject: driver.ThermocyclerDriverV2, connection: AsyncMock
) -> None:
    """It should send a Plate Lift command."""
    await subject.lift_plate()

    expected = CommandBuilder(terminator=driver.TC_COMMAND_TERMINATOR).add_gcode(
        gcode="M128"
    )

    connection.send_command.assert_called_once_with(command=expected, retries=3)


async def test_get_lid_status(
    subject: driver.ThermocyclerDriverV2, connection: AsyncMock
) -> None:
    """It should send a get lid status command and parse response."""
    connection.send_command.return_value = "M119 Lid:open Seal:engaged OK\n"

    response = await subject.get_lid_status()

    expected = CommandBuilder(terminator=driver.TC_COMMAND_TERMINATOR).add_gcode(
        gcode="M119"
    )

    connection.send_command.assert_called_once_with(command=expected, retries=3)
    assert response == ThermocyclerLidStatus.OPEN


@pytest.mark.parametrize(
    argnames=["requested_temp", "actual_temp"],
    argvalues=[
        [driver.LID_TARGET_MIN + 5, driver.LID_TARGET_MIN + 5],
        [driver.LID_TARGET_MIN - 5, driver.LID_TARGET_MIN],
        [driver.LID_TARGET_MAX + 5, driver.LID_TARGET_MAX],
    ],
)
async def test_set_lid_temp(
    subject: driver.ThermocyclerDriverV2,
    connection: AsyncMock,
    requested_temp: float,
    actual_temp: float,
) -> None:
    """It should send a set lid temperature command honoring the min/max."""
    await subject.set_lid_temperature(temp=requested_temp)

    expected = (
        CommandBuilder(terminator=driver.TC_COMMAND_TERMINATOR)
        .add_gcode(gcode="M140")
        .add_float(prefix="S", value=actual_temp, precision=TC_GCODE_ROUNDING_PRECISION)
    )

    connection.send_command.assert_called_once_with(command=expected, retries=3)


async def test_get_lid_temp(
    subject: driver.ThermocyclerDriverV2, connection: AsyncMock
) -> None:
    """It should send a get lid temperature command and parse response."""
    connection.send_command.return_value = "M141 T:100.000 C:22.041 OK\n'"

    response = await subject.get_lid_temperature()

    expected = CommandBuilder(terminator=driver.TC_COMMAND_TERMINATOR).add_gcode(
        gcode="M141"
    )

    connection.send_command.assert_called_once_with(command=expected, retries=3)
    assert response == Temperature(target=100, current=22.04)


@pytest.mark.parametrize(
    argnames=["temp", "hold_time", "volume", "ramp_rate", "expected_body"],
    argvalues=[
        [50, 2, 32, None, "S50 H2 V32"],
        [50, None, None, None, "S50"],
        [50, 2, None, None, "S50 H2"],
        [50, None, 32, None, "S50 V32"],
        [-5, 2, 32, None, "S0 H2 V32"],
        [102, 2, 32, None, "S99 H2 V32"],
        [102, 2, 32, 1.0, "S99 H2 V32 R1.0"],
    ],
)
async def test_set_plate_temp(
    subject: driver.ThermocyclerDriverV2,
    connection: AsyncMock,
    temp: float,
    hold_time: Optional[float],
    volume: Optional[float],
    ramp_rate: Optional[float],
    expected_body: str,
) -> None:
    """It should send a set plate temperature command."""
    await subject.set_plate_temperature(
        temp=temp, hold_time=hold_time, volume=volume, ramp_rate=ramp_rate
    )

    expected = (
        CommandBuilder(terminator=driver.TC_COMMAND_TERMINATOR)
        .add_gcode(gcode="M104")
        .add_element(element=expected_body)
    )

    connection.send_command.assert_called_once_with(command=expected, retries=3)


async def test_get_plate_temp(
    subject: driver.ThermocyclerDriverV2, connection: AsyncMock
) -> None:
    """It should send a command to get the plate temperature and parse the response."""
    connection.send_command.return_value = (
        "M105 T:30.000 C:23.317 H:120 Total_H:120 At_target?:0 OK\n"
    )

    response = await subject.get_plate_temperature()

    expected = CommandBuilder(terminator=driver.TC_COMMAND_TERMINATOR).add_gcode(
        gcode="M105"
    )

    connection.send_command.assert_called_once_with(command=expected, retries=3)
    assert response == PlateTemperature(target=30, current=23.32, hold=120)


async def test_set_ramp_rate(
    subject: driver.ThermocyclerDriver, connection: AsyncMock
) -> None:
    """It should send a set ramp rate command."""
    await subject.set_ramp_rate(ramp_rate=22)

    expected = (
        CommandBuilder(terminator=driver.TC_COMMAND_TERMINATOR)
        .add_gcode(gcode="M566")
        .add_float(prefix="S", value=22, precision=TC_GCODE_ROUNDING_PRECISION)
    )

    connection.send_command.assert_called_once_with(command=expected, retries=3)


async def test_deactivate_all(
    subject: driver.ThermocyclerDriverV2, connection: AsyncMock
) -> None:
    """It should send a deactivate all command."""
    await subject.deactivate_all()

    expected = CommandBuilder(terminator=driver.TC_COMMAND_TERMINATOR).add_gcode(
        gcode="M18"
    )

    connection.send_command.assert_called_once_with(command=expected, retries=3)


async def test_deactivate_lid(
    subject: driver.ThermocyclerDriverV2, connection: AsyncMock
) -> None:
    """It should send a deactivate lid command."""
    await subject.deactivate_lid()

    expected = CommandBuilder(terminator=driver.TC_COMMAND_TERMINATOR).add_gcode(
        gcode="M108"
    )

    connection.send_command.assert_called_once_with(command=expected, retries=3)


async def test_deactivate_block(
    subject: driver.ThermocyclerDriverV2, connection: AsyncMock
) -> None:
    """It should send a deactivate block command."""
    await subject.deactivate_block()

    expected = CommandBuilder(terminator=driver.TC_COMMAND_TERMINATOR).add_gcode(
        gcode="M14"
    )

    connection.send_command.assert_called_once_with(command=expected, retries=3)


async def test_device_info(
    subject: driver.ThermocyclerDriverV2, connection: AsyncMock
) -> None:
    """It should send a get device info command and parse response."""
    connection.send_command.return_value = (
        "M115 FW:(dev)--b4eeff7 HW:Opentrons-thermocycler-gen2 SerialNo:EMPTYSN OK\n"
    )

    device_info = await subject.get_device_info()

    get_device_info = CommandBuilder(terminator=driver.TC_COMMAND_TERMINATOR).add_gcode(
        gcode="M115"
    )
    reset_reason = CommandBuilder(terminator=driver.TC_COMMAND_TERMINATOR).add_gcode(
        gcode="M114"
    )

    connection.send_command.assert_any_call(command=get_device_info, retries=3)
    connection.send_command.assert_called_with(command=reset_reason, retries=3)

    assert device_info == {
        "serial": "EMPTYSN",
        "model": "Opentrons-thermocycler-gen2",
        "version": "(dev)--b4eeff7",
    }


async def test_enter_bootloader(
    subject: driver.ThermocyclerDriverV2, connection: AsyncMock
) -> None:
    """It should send the bootloader command and close the connection"""

    await subject.enter_programming_mode()

    expected = CommandBuilder(terminator=driver.TC_COMMAND_TERMINATOR).add_gcode(
        gcode="dfu"
    )

    connection.send_dfu_command.assert_called_once_with(command=expected)

    assert connection.close.called



from typing import Any, Type, cast

import mock
import pytest

from opentrons_shared_data.labware import load_definition as load_labware_definition
from opentrons_shared_data.module.types import ModuleDefinitionV3

import opentrons.protocol_api as papi
from opentrons.drivers.types import (
    HeaterShakerLabwareLatchStatus,
    ThermocyclerLidStatus,
)
from opentrons.hardware_control import modules as hw_modules
from opentrons.hardware_control.modules.magdeck import OFFSET_TO_LABWARE_BOTTOM
from opentrons.hardware_control.modules.types import (
    SpeedStatus,
    ThermocyclerModuleModel,
)
from opentrons.protocol_api.core.legacy.module_geometry import (
    PipetteMovementRestrictedByHeaterShakerError,
    models_compatible,
)
from opentrons.protocol_api.protocol_context import ProtocolContext
from opentrons.protocols.api_support.deck_type import STANDARD_OT2_DECK
from opentrons.protocols.api_support.types import APIVersion
from opentrons.types import Location, Point


@pytest.fixture
def mock_hardware() -> mock.MagicMock:
    return mock.MagicMock()


@pytest.fixture
def mock_module_controller() -> mock.MagicMock:
    return mock.MagicMock()


@pytest.fixture
def mock_pipette_location() -> mock.MagicMock:
    return mock.MagicMock(return_value=Location(point=Point(1, 2, 3), labware=None))


@pytest.fixture
def ctx_with_tempdeck(
    mock_hardware: mock.MagicMock, mock_module_controller: mock.MagicMock
) -> papi.ProtocolContext:
    """Context fixture with a mock temp deck."""
    mock_module_controller.model.return_value = "temperatureModuleV2"
    mock_module_controller.mock_add_spec(hw_modules.TempDeck)
    mock_hardware.attached_modules = [mock_module_controller]

    return papi.create_protocol_context(
        api_version=APIVersion(2, 13),
        hardware_api=mock_hardware,
        deck_type=STANDARD_OT2_DECK,
    )


@pytest.fixture
def ctx_with_magdeck(
    mock_hardware: mock.AsyncMock, mock_module_controller: mock.MagicMock
) -> papi.ProtocolContext:
    """Context fixture with a mock mag deck."""
    mock_module_controller.model.return_value = "magneticModuleV1"
    mock_module_controller.mock_add_spec(hw_modules.MagDeck)
    mock_hardware.attached_modules = [mock_module_controller]

    return papi.create_protocol_context(
        api_version=APIVersion(2, 13),
        hardware_api=mock_hardware,
        deck_type=STANDARD_OT2_DECK,
    )


@pytest.fixture
async def ctx_with_thermocycler(
    mock_hardware: mock.AsyncMock, mock_module_controller: mock.MagicMock
) -> papi.ProtocolContext:
    """Context fixture with a mock thermocycler."""
    mock_module_controller.model.return_value = "thermocyclerModuleV1"
    mock_module_controller.mock_add_spec(hw_modules.Thermocycler)
    mock_hardware.attached_modules = [mock_module_controller]

    return papi.create_protocol_context(
        api_version=APIVersion(2, 13),
        hardware_api=mock_hardware,
        deck_type=STANDARD_OT2_DECK,
    )


@pytest.fixture
def ctx_with_heater_shaker(
    mock_hardware: mock.AsyncMock,
    mock_module_controller: mock.MagicMock,
    mock_pipette_location: mock.MagicMock,
) -> papi.ProtocolContext:
    """Context fixture with a mock heater-shaker."""
    mock_module_controller.model.return_value = "heaterShakerModuleV1"
    mock_module_controller.mock_add_spec(hw_modules.HeaterShaker)
    mock_hardware.attached_modules = [mock_module_controller]

    ctx = papi.create_protocol_context(
        api_version=APIVersion(2, 13),
        hardware_api=mock_hardware,
        deck_type=STANDARD_OT2_DECK,
    )
    ctx.location_cache = mock_pipette_location
    return ctx


# ______ load_module tests _______


def test_load_module(ctx_with_tempdeck: papi.ProtocolContext) -> None:
    ctx_with_tempdeck.home()
    mod = ctx_with_tempdeck.load_module("tempdeck", 1)
    assert isinstance(mod, papi.TemperatureModuleContext)


def test_load_module_default_slot(ctx_with_thermocycler: papi.ProtocolContext) -> None:
    ctx_with_thermocycler.home()
    mod = ctx_with_thermocycler.load_module("thermocycler")
    assert isinstance(mod, papi.ThermocyclerContext)


def test_no_slot_module_error(ctx_with_magdeck: papi.ProtocolContext) -> None:
    ctx_with_magdeck.home()
    with pytest.raises(ValueError):
        assert ctx_with_magdeck.load_module("magdeck")


def test_invalid_slot_module_error(ctx_with_thermocycler: papi.ProtocolContext) -> None:
    ctx_with_thermocycler.home()
    with pytest.raises(ValueError):
        assert ctx_with_thermocycler.load_module("thermocycler", 1)


def test_bad_slot_module_error(ctx_with_tempdeck: papi.ProtocolContext) -> None:
    ctx_with_tempdeck.home()
    with pytest.raises(ValueError):
        assert ctx_with_tempdeck.load_module("thermocycler", 42)


def test_incorrect_module_error(ctx_with_tempdeck: papi.ProtocolContext) -> None:
    ctx_with_tempdeck.home()
    with pytest.raises(ValueError):
        assert ctx_with_tempdeck.load_module("the cool module", 1)


# TODO(mm, 2023-04-28): This test uses mod.geometry, which was always a quasi-implementation-detail
# and was removed in PAPIv2.14. We should make sure this logic is adequately covered elsewhere
# in a way that doesn't depend on mod.geometry, and then delete this test.
@pytest.mark.apiv2_non_pe_only
@pytest.mark.parametrize(
    "loadname,klass,model",
    [
        ("tempdeck", papi.TemperatureModuleContext, "temperatureModuleV1"),
        ("temperature module", papi.TemperatureModuleContext, "temperatureModuleV1"),
        (
            "temperature module gen2",
            papi.TemperatureModuleContext,
            "temperatureModuleV2",
        ),
        ("magdeck", papi.MagneticModuleContext, "magneticModuleV1"),
        ("magnetic module", papi.MagneticModuleContext, "magneticModuleV1"),
        ("magnetic module gen2", papi.MagneticModuleContext, "magneticModuleV2"),
        ("thermocycler", papi.ThermocyclerContext, "thermocyclerModuleV1"),
        ("thermocycler module", papi.ThermocyclerContext, "thermocyclerModuleV1"),
        ("thermocycler module gen2", papi.ThermocyclerContext, "thermocyclerModuleV2"),
        ("heaterShakerModuleV1", papi.HeaterShakerContext, "heaterShakerModuleV1"),
    ],
)
def test_load_simulating_module(
    ctx: ProtocolContext, loadname: str, klass: Type[papi.ProtocolContext], model: str
) -> None:
    """Check that a known module will not throw an error if in simulation mode.

    Note: This is basically an integration test that checks that a module can be
          loaded correctly. So it checks the `load_module` function all the way through
          module instance creation, which includes fetching module definition, loading
          geometry and finding attached modules or creating simulated module in order to
          finally build an instance of the specified module.
    """
    ctx.home()
    mod = ctx.load_module(loadname, 7)
    assert isinstance(mod, klass)
    assert mod.geometry.model.value == model
    assert mod._core._sync_module_hardware.model() == model


# _________ Magnetic Module tests __________


def test_magdeck(ctx_with_magdeck: papi.ProtocolContext) -> None:
    mod = ctx_with_magdeck.load_module("Magnetic Module", 1)
    assert ctx_with_magdeck.deck[1] == mod.geometry


# _________ Thermocycler tests __________


def test_thermocycler(ctx_with_thermocycler: papi.ProtocolContext) -> None:
    mod = ctx_with_thermocycler.load_module("thermocycler")
    assert ctx_with_thermocycler.deck[7] == mod.geometry


def test_thermocycler_lid_status(
    ctx_with_thermocycler: papi.ProtocolContext, mock_module_controller: mock.MagicMock
) -> None:
    mod = ctx_with_thermocycler.load_module("thermocycler")
    m = mock.PropertyMock(return_value=ThermocyclerLidStatus.OPEN)
    assert isinstance(mod, papi.ThermocyclerContext)
    type(mock_module_controller).lid_status = m
    assert mod.lid_position == "open"


def test_thermocycler_profile_invalid_repetitions(
    ctx_with_thermocycler: papi.ProtocolContext, mock_module_controller: mock.MagicMock
) -> None:
    mod = ctx_with_thermocycler.load_module("thermocycler")
    assert isinstance(mod, papi.ThermocyclerContext)

    with pytest.raises(ValueError, match="positive integer"):
        mod.execute_profile(
            steps=[
                {"temperature": 10, "hold_time_seconds": 30},
                {"temperature": 30, "hold_time_seconds": 90},
            ],
            repetitions=-1,
        )


def test_thermocycler_profile_no_temperature(
    ctx_with_thermocycler: papi.ProtocolContext, mock_module_controller: mock.MagicMock
) -> None:
    mod = ctx_with_thermocycler.load_module("thermocycler")
    assert isinstance(mod, papi.ThermocyclerContext)
    with pytest.raises(ValueError, match="temperature must be defined"):
        mod.execute_profile(
            steps=[
                {"temperature": 10, "hold_time_seconds": 30},
                {"hold_time_seconds": 90},  # type: ignore
            ],
            repetitions=5,
        )


def test_thermocycler_profile_no_hold(
    ctx_with_thermocycler: papi.ProtocolContext, mock_module_controller: mock.MagicMock
) -> None:
    mod = ctx_with_thermocycler.load_module("thermocycler")
    assert isinstance(mod, papi.ThermocyclerContext)
    with pytest.raises(
        ValueError, match="either hold_time_minutes or hold_time_seconds"
    ):
        mod.execute_profile(
            steps=[{"temperature": 10, "hold_time_seconds": 30}, {"temperature": 30}],
            repetitions=5,
        )


@pytest.mark.apiv2_non_pe_only  # Semi plate configuration removed in PAPIv2.14.
def test_thermocycler_semi_plate_configuration(ctx: papi.ProtocolContext) -> None:
    labware_name = "nest_96_wellplate_100ul_pcr_full_skirt"
    mod = ctx.load_module("thermocycler", configuration="semi")
    assert mod.geometry.labware_offset == Point(-23.28, 82.56, 97.8)

    tc_labware = mod.load_labware(labware_name)

    other_labware = ctx.load_labware(labware_name, 2)
    without_first_two_cols = other_labware.wells()[16::]
    for tc_well, other_well in zip(tc_labware.wells(), without_first_two_cols):
        tc_well_name = tc_well.display_name.split()[0]
        other_well_name = other_well.display_name.split()[0]
        assert tc_well_name == other_well_name


def test_thermocycler_flag_unsafe_move(
    ctx_with_thermocycler: papi.ProtocolContext, mock_module_controller: mock.MagicMock
) -> None:
    """Flag unsafe should raise if the lid is open and source or target is
    the labware on thermocycler."""
    mod = ctx_with_thermocycler.load_module("thermocycler", configuration="semi")
    labware_name = "nest_96_wellplate_100ul_pcr_full_skirt"
    tc_labware = mod.load_labware(labware_name)

    with_tc_labware = Location(None, tc_labware)  # type: ignore[arg-type]
    without_tc_labware = Location(None, None)  # type: ignore[arg-type]

    m = mock.PropertyMock(return_value=ThermocyclerLidStatus.CLOSED)
    type(mock_module_controller).lid_status = m

    assert hasattr(mod._core, "flag_unsafe_move")

    with pytest.raises(RuntimeError, match="Cannot move to labware"):
        mod._core.flag_unsafe_move(with_tc_labware, without_tc_labware)
    with pytest.raises(RuntimeError, match="Cannot move to labware"):
        mod._core.flag_unsafe_move(without_tc_labware, with_tc_labware)


# __________ Heater Shaker tests __________


@pytest.mark.parametrize(
    argnames=["labware_name", "is_tiprack"],
    argvalues=[("geb_96_tiprack_1000ul", True), ("biorad_384_wellplate_50ul", False)],
)
def test_heater_shaker_unsafe_move_flagger(
    ctx_with_heater_shaker: papi.ProtocolContext,
    mock_module_controller: mock.MagicMock,
    labware_name: str,
    is_tiprack: bool,
) -> None:
    """It should call unsafe movement flagger with correct args."""
    mock_speed_status = mock.PropertyMock(return_value=SpeedStatus.DECELERATING)
    mock_latch_status = mock.PropertyMock(
        return_value=HeaterShakerLabwareLatchStatus.IDLE_CLOSED
    )
    type(mock_module_controller).speed_status = mock_speed_status
    type(mock_module_controller).labware_latch_status = mock_latch_status

    mod = ctx_with_heater_shaker.load_module("heaterShakerModuleV1", 3)
    assert isinstance(mod, papi.HeaterShakerContext)

    labware = ctx_with_heater_shaker.load_labware(labware_name, 5)

    mod._core.geometry.flag_unsafe_move = mock.MagicMock()  # type: ignore[attr-defined]

    mod._core.flag_unsafe_move(to_loc=labware.wells()[1].top(), is_multichannel=False)  # type: ignore[attr-defined]

    mod._core.geometry.flag_unsafe_move.assert_called_once_with(  # type: ignore[attr-defined]
        to_slot=5,
        is_tiprack=is_tiprack,
        is_using_multichannel=False,
        is_labware_latch_closed=True,
        is_plate_shaking=True,
    )


def test_hs_flag_unsafe_move_raises(
    ctx_with_heater_shaker: papi.ProtocolContext,
) -> None:
    """Test unsafe move raises underlying error."""

    def raiser(*args: Any, **kwargs: Any) -> None:
        raise PipetteMovementRestrictedByHeaterShakerError("uh oh")

    mod = ctx_with_heater_shaker.load_module("heaterShakerModuleV1", 3)
    labware = ctx_with_heater_shaker.load_labware("geb_96_tiprack_1000ul", 5)

    assert isinstance(mod, papi.HeaterShakerContext)
    mod._core.geometry.flag_unsafe_move = mock.MagicMock(side_effect=raiser)  # type: ignore[attr-defined]

    with pytest.raises(PipetteMovementRestrictedByHeaterShakerError, match="uh oh"):
        mod._core.flag_unsafe_move(  # type: ignore[attr-defined]
            to_loc=labware.wells()[1].top(), is_multichannel=False
        )


def test_hs_flag_unsafe_move_skips_non_labware_locations(
    ctx_with_heater_shaker: papi.ProtocolContext,
    mock_module_controller: mock.MagicMock,
) -> None:
    """Test that purely point locations do not raise error."""
    mod = ctx_with_heater_shaker.load_module("heaterShakerModuleV1", 3)
    assert isinstance(mod, papi.HeaterShakerContext)
    mod._core.geometry.flag_unsafe_move = mock.MagicMock()  # type: ignore[attr-defined]

    mod._core.flag_unsafe_move(  # type: ignore[attr-defined]
        to_loc=Location(point=Point(1, 2, 3), labware=None), is_multichannel=False
    )
    mod._core.geometry.flag_unsafe_move.assert_not_called()  # type: ignore[attr-defined]


def test_heater_shaker_loading(
    ctx_with_heater_shaker: papi.ProtocolContext,
    mock_module_controller: mock.MagicMock,
) -> None:
    """It should load a heater-shaker in the specified slot."""
    mod = ctx_with_heater_shaker.load_module("heaterShakerModuleV1", 3)
    assert ctx_with_heater_shaker.deck[3] == mod.geometry


# __________ Testing loading Labware on modules ___________


def test_module_load_labware(ctx_with_tempdeck: papi.ProtocolContext) -> None:
    labware_name = "corning_96_wellplate_360ul_flat"
    # TODO Ian 2019-05-29 load fixtures, not real defs
    labware_def = load_labware_definition(labware_name, 1)
    mod = ctx_with_tempdeck.load_module("Temperature Module", 1)
    assert mod.labware is None
    lw = mod.load_labware(labware_name)
    assert (
        labware_def["schemaVersion"] == 2
    )  # For the presence of ["cornerOffsetFromSlot"].
    lw_offset = Point(
        labware_def["cornerOffsetFromSlot"]["x"],
        labware_def["cornerOffsetFromSlot"]["y"],
        labware_def["cornerOffsetFromSlot"]["z"],
    )
    assert lw._core.get_geometry().offset == lw_offset + mod.geometry.location.point  # type: ignore[attr-defined]
    assert lw.name == labware_name


def test_module_load_labware_with_label(
    ctx_with_tempdeck: papi.ProtocolContext,
) -> None:
    labware_name = "corning_96_wellplate_360ul_flat"
    mod = ctx_with_tempdeck.load_module("Temperature Module", 1)
    lw = mod.load_labware(labware_name, label="my cool labware")
    assert lw.name == "my cool labware"


def test_module_load_invalid_labware(ctx_with_tempdeck: papi.ProtocolContext) -> None:
    labware_name = "corning_96_wellplate_360ul_flat"
    mod = ctx_with_tempdeck.load_module("Temperature Module", 1)
    # wrong version number
    with pytest.raises(FileNotFoundError):
        mod.load_labware(labware_name, namespace="opentrons", version=100)
    # wrong namespace
    with pytest.raises(FileNotFoundError):
        mod.load_labware(labware_name, namespace="fake namespace", version=1)
    # valid info
    assert mod.load_labware(labware_name, namespace="opentrons", version=1)


def test_deprecated_module_load_labware_by_name(
    ctx_with_tempdeck: papi.ProtocolContext,
) -> None:
    """It should call load labware"""
    mod = ctx_with_tempdeck.load_module("Temperature Module", 1)
    with mock.patch.object(mod, "load_labware") as mock_load_labware:
        mod.load_labware_by_name(
            name="a module", namespace="ns", label="a label", version=2
        )
        mock_load_labware.assert_called_once_with(
            name="a module", label="a label", namespace="ns", version=2
        )


@pytest.mark.apiv2_non_pe_only  # engage(height=...) param was removed in PAPIv2.14.
async def test_magdeck_gen1_labware_props(ctx: papi.ProtocolContext) -> None:
    # TODO Ian 2019-05-29 load fixtures, not real defs
    labware_name = "biorad_96_wellplate_200ul_pcr"
    labware_def = load_labware_definition(labware_name, 1)
    mod = ctx.load_module("magdeck", 1)
    assert mod.labware is None
    assert isinstance(mod, papi.MagneticModuleContext)
    mod.engage(height=45)
    assert mod._core._sync_module_hardware.current_height == 45  # type: ignore[attr-defined]
    with pytest.raises(ValueError):
        mod.engage(height=45.1)  # max engage height for gen1 is 45 mm
    mod.load_labware(labware_name)
    mod.engage()
    if "magneticModuleEngageHeight" not in labware_def["parameters"]:
        assert False
    lw_offset = labware_def["parameters"]["magneticModuleEngageHeight"]
    assert await mod._core._sync_module_hardware._driver.get_plate_height() == lw_offset  # type: ignore[attr-defined]
    mod.disengage()
    mod.engage(offset=2)
    assert (
        await mod._core._sync_module_hardware._driver.get_plate_height()  # type: ignore[attr-defined]
        == lw_offset + 2
    )
    mod.disengage()
    mod.engage(height=3)
    assert await mod._core._sync_module_hardware._driver.get_plate_height() == 3  # type: ignore[attr-defined]
    mod.geometry.reset_labware()
    labware_name = "corning_96_wellplate_360ul_flat"
    mod.load_labware(labware_name)
    with pytest.raises(ValueError):
        mod.engage()
    with pytest.raises(ValueError):
        mod.engage(offset=1)
    mod.engage(height=2)
    assert await mod._core._sync_module_hardware._driver.get_plate_height() == 2  # type: ignore[attr-defined]
    mod.engage(height=0)
    assert await mod._core._sync_module_hardware._driver.get_plate_height() == 0  # type: ignore[attr-defined]
    mod.engage(height_from_base=2)
    assert (
        await mod._core._sync_module_hardware._driver.get_plate_height()  # type: ignore[attr-defined]
        == 2 + OFFSET_TO_LABWARE_BOTTOM[mod._core._sync_module_hardware.model()]  # type: ignore[attr-defined]
    )


@pytest.mark.apiv2_non_pe_only  # engage(height=...) param was removed in PAPIv2.14.
def test_magdeck_gen2_labware_props(ctx: papi.ProtocolContext) -> None:
    mod = ctx.load_module("magnetic module gen2", 1)
    assert isinstance(mod, papi.MagneticModuleContext)
    mod.engage(height=25)
    assert mod._core._sync_module_hardware.current_height == 25  # type: ignore[attr-defined]
    with pytest.raises(ValueError):
        mod.engage(height=25.1)  # max engage height for gen2 is 25 mm
    mod.engage(height=0)
    assert mod._core._sync_module_hardware.current_height == 0  # type: ignore[attr-defined]


def test_module_compatibility() -> None:
    assert (
        models_compatible(
            requested_model=ThermocyclerModuleModel["THERMOCYCLER_V1"],
            candidate_definition=cast(
                ModuleDefinitionV3, {"model": "thermocyclerModuleV1"}
            ),
        )
        is True
    )

    assert (
        models_compatible(
            requested_model=ThermocyclerModuleModel["THERMOCYCLER_V2"],
            candidate_definition=cast(
                ModuleDefinitionV3,
                {
                    "model": "thermocyclerModuleV1",
                    "compatibleWith": ["thermocyclerModuleV2"],
                },
            ),
        )
        is True
    )

    assert (
        models_compatible(
            requested_model=ThermocyclerModuleModel["THERMOCYCLER_V1"],
            candidate_definition=cast(
                ModuleDefinitionV3,
                {
                    "model": "thermocyclerModuleV2",
                    "compatibleWith": [],
                },
            ),
        )
        is False
    )
